import logging
import ibmsecurity.isam.base.container_ext.container as container

logger = logging.getLogger(__name__)

uri = "/isam/container_ext/health"

requires_version = "10.0.7.0"
requires_model = "Appliance"


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving health for all (running?) containers TODO: THIS FAILS IN 10.0.8
    ./testisam_cmd.py --hostname=${ISAM_LMI} --method=ibmsecurity.isam.base.container_ext.health.get_all

    """
    return isamAppliance.invoke_get(
        "Retrieving health for all containers",
        f"{uri}",
        requires_model=requires_model,
        requires_version=requires_version,
    )


def get(isamAppliance, name=None, container_id=None, check_mode=False, force=False):
    """
    Get health for a specific container TODO: THIS FAILS IN 10.0.8
    Supply either name or id.
    ./testisam_cmd.py --hostname=${ISAM_LMI} --method=ibmsecurity.isam.base.container_ext.health.get --method_options="container_id=a6252734-87d3-11ef-bec6-000c29d80c72"
    """
    if container_id is None and name is not None:
        container_id = container.search(isamAppliance, name)
        container_id = container_id.get('data', None)

    if container_id is not None:
        return isamAppliance.invoke_get(
            "Retrieving health for container",
            f"{uri}/{container_id}",
            requires_model=requires_model,
            requires_version=requires_version,
        )
    else:
        return isamAppliance.create_return_object()



import logging
import ibmsecurity.utilities.tools

try:
    basestring
except NameError:
    basestring = (str, bytes)

logger = logging.getLogger(__name__)

uri = "/isam/cluster/v2"
requires_modules = None
requires_version = "9.0.4.0"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve the current service configuration
    """
    return isamAppliance.invoke_get("Retrieve the current service configuration", uri,
                                    requires_modules=requires_modules, requires_version=requires_version)


def set(isamAppliance, hvdb_db_type=None, hvdb_address=None, hvdb_port=None, hvdb_user=None,
        hvdb_password=None, hvdb_db2_alt_address=None, hvdb_db2_alt_port=None, hvdb_db_name=None, hvdb_db_secure=None,
        hvdb_driver_type=None, hvdb_solid_tc=None, check_mode=False, force=False):
    """
    Set service configuration
    """
    warnings = []
    # Create a simple json with just the main client attributes
    service_json = {}
    # Add attributes that have been supplied... otherwise skip them.
    if hvdb_db_type is not None:
        service_json["hvdb_db_type"] = hvdb_db_type
    if hvdb_address is not None:
        service_json["hvdb_address"] = hvdb_address
    if hvdb_port is not None:
        service_json["hvdb_port"] = hvdb_port
    if hvdb_user is not None:
        service_json["hvdb_user"] = hvdb_user
    if hvdb_password is not None:
        warnings.append("Since existing hvdb_password cannot be read - this call will not be idempotent.")
        service_json["hvdb_password"] = hvdb_password
    if hvdb_db2_alt_address is not None:
        service_json["hvdb_db2_alt_address"] = hvdb_db2_alt_address
    if hvdb_db2_alt_port is not None:
        service_json["hvdb_db2_alt_port"] = hvdb_db2_alt_port
    if hvdb_db_name is not None:
        service_json["hvdb_db_name"] = hvdb_db_name
    if hvdb_db_secure is not None:
        service_json["hvdb_db_secure"] = hvdb_db_secure
    if hvdb_driver_type is not None:
        service_json["hvdb_driver_type"] = hvdb_driver_type
    if hvdb_solid_tc is not None:
        if (isinstance(hvdb_solid_tc, basestring)):
            import ast
            hvdb_solid_tc = ast.literal_eval(hvdb_solid_tc)
            service_json["hvdb_solid_tc"] = hvdb_solid_tc

    if force is True or _check(isamAppliance, service_json) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post("Set service configuration", uri, service_json,
                                             requires_modules=requires_modules, requires_version=requires_version,
                                             warnings=warnings)

    return isamAppliance.create_return_object()


def _check(isamAppliance, service_json):
    """
    Check if provided json values match the configuration on appliance

    :param isamAppliance:
    :param service_json:
    :return:
    """
    ret_obj = get(isamAppliance)
    logger.debug("Appliance current configuration: {0}".format(ret_obj['data']))
    logger.debug("JSON to Apply: {0}".format(service_json))

    for key, value in service_json.items():
        try:
            if isinstance(value, list):
                if ibmsecurity.utilities.tools.json_sort(
                        ret_obj['data'][key] != ibmsecurity.utilities.tools.json_sort(value)):
                    logger.debug(
                        "For key: {0}, values: {1} and {2} do not match.".format(key, value, ret_obj['data'][key]))
                    return False
            else:
                if ret_obj['data'][key] != value:
                    logger.debug(
                        "For key: {0}, values: {1} and {2} do not match.".format(key, value, ret_obj['data'][key]))
                    return False
        except:  # In case there is an error looking up the key in existing configuration (missing)
            logger.debug("Exception processing Key: {0} Value: {1} - missing key in current config?".format(key, value))
            return False

    logger.debug("JSON provided already is contained in current appliance configuration.")
    return True


def compare(isamAppliance1, isamAppliance2):
    """
    Compare service configuration between two containers
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
import ibmsecurity.utilities.tools
import ibmsecurity.isam.base.network.interfaces

logger = logging.getLogger(__name__)


def add(isamAppliance, label, address, prefixLength, vlanId=None, allowManagement=False,
        enabled=True, check_mode=False,
        force=False):
    """
    Adding an IPv6 address to an interface
    """
    add_needed = True
    ret_obj = {}
    warnings = []
    if force is False:
        ret_obj, warnings = ibmsecurity.isam.base.network.interfaces._get_interface(isamAppliance, label, vlanId)
        if ret_obj is not None:
            del ret_obj['objType']
            del ret_obj['type']
            for addr in ret_obj['ipv6']['addresses']:
                if addr['address'] == address:
                    add_needed = False
                    break
        else:
            warnings.append("Interface {0} not found, Add is not supported.".format(label))
            return isamAppliance.create_return_object(changed=False, warnings=warnings)

    # Cannot add an address and keep dhcp enabled
    if add_needed is True and ret_obj['ipv6']['dhcp']['enabled'] is True:
        ret_obj['ipv6']['dhcp']['enabled'] = False

    if force is True or add_needed is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            addr = {
                '_isNew': True,
                'objType': 'ipv6Address',
                'uuid': None,
                'address': address,
                'prefixLength': prefixLength,
                'maskOrPrefix': '',
                'allowManagement': allowManagement,
                'enabledAddress': True,
                'enabled': enabled
            }
            ret_obj['ipv6']['addresses'].append(addr)
            return ibmsecurity.isam.base.network.interfaces._update_interface(isamAppliance, ret_obj)

    return isamAppliance.create_return_object()


def delete(isamAppliance, label, address, vlanId=None, check_mode=False, force=False):
    """
    Deleting an IPv6 address from an interface
    """
    delete_needed = False
    ret_obj = {}
    warnings = []
    if force is False:
        ret_obj, warnings = ibmsecurity.isam.base.network.interfaces._get_interface(isamAppliance, label, vlanId)
        if ret_obj is not None:
            for addr in ret_obj['ipv6']['addresses']:
                if addr['address'] == address:
                    delete_needed = True
                    ret_obj['ipv6']['addresses'].remove(addr)
                    break
        else:
            warnings.append("Interface {0} not found, Delete is not supported.".format(label))
            return isamAppliance.create_return_object(changed=False, warnings=warnings)

    if force is True or delete_needed is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return ibmsecurity.isam.base.network.interfaces._update_interface(isamAppliance, ret_obj)

    return isamAppliance.create_return_object()


def update(isamAppliance, label, address, new_address, prefixLength, vlanId=None, allowManagement=False, enabled=True,
           check_mode=False, force=False):
    """
    Updating an IPv6 address on an interface
    """
    update_needed = False
    ret_obj = {}
    warnings = []
    if force is False:
        ret_obj, warnings = ibmsecurity.isam.base.network.interfaces._get_interface(isamAppliance, label, vlanId)
        if ret_obj is not None:
            for addr in ret_obj['ipv6']['addresses']:
                if addr['address'] == address:
                    upd_addr = {
                        'uuid': addr['uuid'],
                        'address': new_address,
                        'prefixLength': prefixLength,
                        'maskOrPrefix': '',
                        'allowManagement': allowManagement,
                        'enabled': enabled
                    }
                    update_needed = not (
                            ibmsecurity.utilities.tools.json_sort(addr) == ibmsecurity.utilities.tools.json_sort(
                        upd_addr))
                    if (update_needed is True):
                        ret_obj['ipv6']['addresses'].remove(addr)
                        ret_obj['ipv6']['addresses'].append(upd_addr)
                    break
        else:
            warnings.append("Interface {0} not found, Update is not supported.".format(label))
            return isamAppliance.create_return_object(changed=False, warnings=warnings)

    if force is True or update_needed is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return ibmsecurity.isam.base.network.interfaces._update_interface(isamAppliance, ret_obj)

    return isamAppliance.create_return_object()


def set_dhcp(isamAppliance, label, vlanId=None, enabled=False, allowManagement=False, check_mode=False, force=False):
    """
    Updating IPv6 dhcp on an interface
    """
    update_needed = False
    ret_obj = {}
    warnings = []
    if force is False:
        ret_obj, warnings = ibmsecurity.isam.base.network.interfaces._get_interface(isamAppliance, label, vlanId)
        if ret_obj is not None:
            upd_dhcp = {
                'enabled': enabled,
                'allowManagement': allowManagement
            }
            update_needed = not (
                    ibmsecurity.utilities.tools.json_sort(
                        ret_obj['ipv6']['dhcp']) == ibmsecurity.utilities.tools.json_sort(
                upd_dhcp))
            if (update_needed is True):
                ret_obj['ipv6']['dhcp'] = upd_dhcp
                # Cannot have DHCP and ip addresses at same time
                if enabled is True:
                    ret_obj['ipv6']['addresses'] = []
        else:
            warnings.append("Interface {0} not found, Set dhcp is not supported.".format(label))
            return isamAppliance.create_return_object(changed=False, warnings=warnings)

    if force is True or update_needed is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return ibmsecurity.isam.base.network.interfaces._update_interface(isamAppliance, ret_obj)

    return isamAppliance.create_return_object()



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get(isamAppliance, check_mode=False, force=False):
    """
    Get management authentication
    """
    return isamAppliance.invoke_get("Get management authentication",
                                    "/isam/management_authentication/")


def set(isamAppliance,
        ldap_host=None,
        ldap_port=None,
        base_dn=None,
        admin_group_dn=None,
        type='remote',
        enable_ssl=False,
        key_database=None,
        cert_label=None,
        user_attribute='uid',
        group_member_attribute='member',
        anon_bind=True,
        bind_dn=None,
        bind_password=None,
        ldap_debug=None,
        enable_usermapping=False,
        usermapping_script=None,
        check_mode=False,
        force=False):
    """
    Set management authentication to remote
    """
    warnings = []
    update_required = False
    json_data = {
        'type': type
    }
    if ldap_host is not None:
        json_data["ldap_host"] = ldap_host
    if ldap_port is not None:
        json_data["ldap_port"] = ldap_port
    if enable_ssl is not None:
        json_data["enable_ssl"] = enable_ssl
    if key_database is not None:
        json_data["key_database"] = key_database
    if cert_label is not None:
        json_data["cert_label"] = cert_label
    if user_attribute is not None:
        json_data["user_attribute"] = user_attribute
    if group_member_attribute is not None:
        json_data["group_member_attribute"] = group_member_attribute
    if base_dn is not None:
        json_data["base_dn"] = base_dn
    if admin_group_dn is not None:
        json_data["admin_group_dn"] = admin_group_dn
    if anon_bind is not None:
        json_data["anon_bind"] = anon_bind
    if bind_dn is not None:
        json_data["bind_dn"] = bind_dn
    if bind_password is not None:
        json_data["bind_password"] = bind_password
    if ldap_debug is not None or ldap_debug == '':
        if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
            warnings.append(
                "Appliance at version: {0}, ldap_debug: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring ldap_debug for this call.".format(
                    isamAppliance.facts["version"], ldap_debug))
        else:
            json_data["ldap_debug"] = ldap_debug
    elif ibmsecurity.utilities.tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") >= 0:
        json_data["ldap_debug"] = False
    if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts["version"], "10.0.2.0") >= 0:
        json_data["enable_usermapping"] = enable_usermapping
        if usermapping_script is not None and enable_usermapping:
            json_data["usermapping_script"] = usermapping_script
    if force is False:
        if bind_password is not None:
            warnings.append("Unable to read existing bind password to check idempotency.")
            update_required = True
        else:
            ret_obj = get(isamAppliance)
            if "bind_dn" in ret_obj['data']:
                if ret_obj["data"]["bind_dn"] is None:
                    del ret_obj["data"]["bind_dn"]
            sorted_json_data = ibmsecurity.utilities.tools.json_sort(json_data)
            logger.debug("Sorted input: {0}".format(sorted_json_data))
            sorted_ret_obj = ibmsecurity.utilities.tools.json_sort(ret_obj['data'])
            logger.debug("Sorted existing data: {0}".format(sorted_ret_obj))
            if sorted_ret_obj != sorted_json_data:
                logger.info("Changes detected, update needed.")
                update_required = True

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Set management authentication to remote",
                "/isam/management_authentication/",
                json_data, warnings=warnings)

    return isamAppliance.create_return_object()


def disable(isamAppliance, check_mode=False, force=False):
    """
    Disable remote management authentication
    """
    return set(isamAppliance=isamAppliance, ldap_host=None, ldap_port=None, base_dn=None, admin_group_dn=None,
               type='local', enable_ssl=None, key_database=None, cert_label=None, user_attribute=None,
               group_member_attribute=None, anon_bind=None, bind_dn=None, bind_password=None, check_mode=check_mode,
               force=force)


def test(isamAppliance, userid, password, check_mode=False, force=False):
    """
    Testing the management authentication
    """
    ret_obj = isamAppliance.invoke_post("Testing the management authentication",
                                        "/isam/management_authentication/",
                                           {
                                            'user': userid,
                                            'password': password
                                           }
                                       )
    ret_obj['changed'] = False

    return ret_obj


def compare(isamAppliance1, isamAppliance2):
    """
    Compare management authentication settings
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

module_uri = "/isam/felb/configuration/attributes"
requires_modules = None
requires_version = None
requires_model = "Appliance"


def add(isamAppliance, name, value, check_mode=False, force=False):
    """
    Creating an attribute
    """

    check_value, warnings = _check(isamAppliance, name, value)

    if force is True or check_value is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post("Creating an attribute", module_uri,
                                             {
                                                 "name": name,
                                                 "value": value
                                             }, requires_modules=requires_modules, requires_version=requires_version,
                                             requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, attribute_name, check_mode=False, force=False):
    """
    Deleting an attribute
    """

    check_obj, warnings = search(isamAppliance, attribute_name)
    if force or check_obj is not False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_delete("Deleting an attribute", "{0}/{1}".format(module_uri, attribute_name),
                                               requires_modules=requires_modules, requires_version=requires_version,
                                               requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def get(isamAppliance, attribute_name):
    """
    Retrieving an attribute
    """
    return isamAppliance.invoke_get("Retrieving an attribute", "{0}/{1}".format(module_uri, attribute_name),
                                    requires_modules=requires_modules, requires_version=requires_version,
                                    requires_model=requires_model)


def get_all(isamAppliance):
    """
    Retrieving attribute names
    """
    return isamAppliance.invoke_get("Retrieving attribute names", "{0}/".format(module_uri),
                                    requires_modules=requires_modules, requires_version=requires_version,
                                    requires_model=requires_model)


def update(isamAppliance, attribute_name, value, check_mode=False, force=False):
    """
    Updating an attribute
    """
    check_value, warnings = _check(isamAppliance, attribute_name, value)

    if force is True or check_value is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put("Updating an attribute", "{0}/{1}".format(module_uri, attribute_name),
                                            {
                                                "value": value
                                            }, requires_version=requires_version, requires_modules=requires_modules,
                                            requires_model=requires_model)

    return isamAppliance.create_return_object(changed=False, warnings=warnings)


def _check(isamAppliance, attribute_name, attribute_value):
    """
    Check for idempotency
    """
    # Error handling to see if attribute exist returns True if attribute doesnt exist
    check_value = False
    warnings = ""

    try:
        temp_obj = get(isamAppliance, attribute_name)
        warnings = temp_obj['warnings']
    except:
        check_value = True
        return check_value, warnings

    if 'value' in temp_obj['data']:
        if temp_obj['data']['value'] != attribute_value:
            check_value = True
            return check_value, warnings
        else:
            check_value = False
            return check_value, warnings
    else:
        check_value = False
        return check_value, warnings


def search(isamAppliance, attribute_name):
    """
    Check for idempotency
    """
    # Error handling to see if attribute exist returns True if attribute doesnt exist

    ret_obj = False
    warnings = ""

    try:
        ret_obj = get(isamAppliance, attribute_name)
        warnings = ret_obj['warnings']
        return ret_obj, warnings
    except:
        ret_obj = False
        return ret_obj, warnings


def compare(isamAppliance1, isamAppliance2):
    """
    Compare between two appliances
    """

    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    obj1_warnings = ret_obj1['warnings']
    obj2_warnings = ret_obj2['warnings']
    warnings = []

    if "Docker" in obj1_warnings or "Docker" in obj2_warnings:
        warnings.append(obj1_warnings)
        warnings.append(obj2_warnings)
        return isamAppliance1.create_return_object(changed=False, warnings=warnings)

    obj1 = {'rc': 0, 'data': [], 'warnings': []}
    obj2 = {'rc': 0, 'data': [], 'warnings': []}

    for attr in ret_obj1["data"]:
        search_result, warnings = search(isamAppliance=isamAppliance1, attribute_name=attr["name"])
        if search_result is not False:
            value = get(isamAppliance1, attribute_name=attr["name"])
            obj1['data'].append({attr["name"]: value["data"]["value"]})
            obj1['warnings'].append(warnings)

    for attr in ret_obj2["data"]:
        search_result, warnings = search(isamAppliance=isamAppliance2, attribute_name=attr["name"])
        if search_result is not False:
            value = get(isamAppliance2, attribute_name=attr["name"])
            obj2['data'].append({attr["name"]: value["data"]["value"]})
            obj2['warnings'].append(warnings)

    return tools.json_compare(obj1, obj2)



import logging
from ibmsecurity.appliance.ibmappliance import IBMError
logger = logging.getLogger(__name__)

uri = "/isam/container_ext/volume"

requires_version = "10.0.7.0"
requires_model = "Appliance"


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving all known deployment properties for volumes
    """
    return isamAppliance.invoke_get(
        "Retrieving container volume properties",
        uri,
        requires_model=requires_model,
        requires_version=requires_version,
    )


def get(isamAppliance, volume_name=None, volume_id=None, filename=None, check_mode=False, force=False):
    """
    Export volume
    """
    import os.path

    if volume_id is None:
        volume_id = search(isamAppliance, volume_name)
        volume_id = volume_id.get('data', None)

    if volume_id is None and volume_name is None:
        return isamAppliance.create_return_object(warnings=["Volume name or volume id need to be specified"])

    if filename is None:
        if volume_name is None:
            filename = f"{volume_id}.zip"
        else:
            filename = f"{volume_name}.zip"
    if force or (not os.path.exists(filename)):
        try:
            return isamAppliance.invoke_get_file(
                "Retrieving configuration for volume",
                f"{uri}/{volume_id}",
                filename,
                requires_model=requires_model,
                requires_version=requires_version,
            )
        except IBMError as e:
            return isamAppliance.create_return_object(warnings=[str(e)])

    return isamAppliance.create_return_object(warnings=[f"File {filename} exists already?"])


# Alias
export = get


def import_zip(isamAppliance, filename, volume_name=None, volume_id=None, check_mode=False, force=False, warnings=[]):
    """
    Import volume

    volume_id - uuid of the volume (the documentation is wrong)
    """
    import os.path
    logger.debug(f"\n\nTrying to import {filename}\n\n")

    if volume_id is None:
        volume_id = search(isamAppliance, volume_name)
        volume_id = volume_id.get('data', None)

    if volume_id is None and volume_name is None:
        return isamAppliance.create_return_object(warnings=["Volume name or volume id need to be specified"])

    if force or os.path.exists(filename):
        return isamAppliance.invoke_put_files(
            "Import a volume zip",
            f"{uri}/{volume_id}",
            [
                {
                    "file_formfield": "volume",
                    "filename": filename,
                    "mimetype": "application/octet-stream",
                }
            ],
            {},
            requires_version=requires_version,
            warnings=warnings,
            requires_model=requires_model,
        )
    return isamAppliance.create_return_object(warnings)


def search(isamAppliance, volume_name, check_mode=False, force=False):
    """
    Return the id of the volume

    :param isamAppliance:
    :param name:
    :param check_mode:
    :param force:
    :return:
    """
    ret_obj = get_all(isamAppliance, check_mode, force)
    return_obj = isamAppliance.create_return_object()
    return_obj["data"] = None
    for obj in ret_obj["data"]:
        if obj["name"] == volume_name:
            return_obj["data"] = obj["id"]
            return_obj["rc"] = 0
            break

    return return_obj


def add(isamAppliance, name, check_mode=False, force=False, warnings=None):
    """
    Add a volume
    """
    if force or not _check(isamAppliance, name):
        if check_mode:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            post_data = {"name": name}

            return isamAppliance.invoke_post(
                "Create new volume",
                uri,
                post_data,
                requires_model=requires_model,
                requires_version=requires_version,
            )

    return isamAppliance.create_return_object()


def _check(isamAppliance, name):
    """
    Check if there's a volume going by name
    """
    ret_obj = get_all(isamAppliance)
    for c in ret_obj["data"]:
        if c.get("name") == name:
            logger.debug(f"Volume {name} exists")
            return True
    return False



import logging

logger = logging.getLogger(__name__)

uri = "/isam/container_ext/metadata"

requires_version = "10.0.7.0"
requires_model = "Appliance"


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving all metadata
    ./testisam_cmd.py --hostname=${ISAM_LMI} --method=ibmsecurity.isam.base.container_ext.metadata.get_all

    """
    return isamAppliance.invoke_get(
        "Retrieving metadata",
        f"{uri}",
        requires_model=requires_model,
        requires_version=requires_version,
    )


def get(isamAppliance, metadata_name, check_mode=False, force=False):
    """
    Get specific metadata for a specific container
    ./testisam_cmd.py --hostname=${ISAM_LMI} --method=ibmsecurity.isam.base.container_ext.metadata.get --method_options="metadata_name=ibm-application-gateway"
    """
    return isamAppliance.invoke_get(
        "Retrieving metadata",
        f"{uri}/{metadata_name}",
        requires_model=requires_model,
        requires_version=requires_version,
    )



import logging
import ibmsecurity.isam.aac.server_connections.ws
from ibmsecurity.utilities.tools import json_sort

logger = logging.getLogger(__name__)

uri = "/mga/scim/configuration"
requires_modules = ["mga", "federation"]
requires_version = "9.0.2"

def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve configuration of SCIM ISAM user settings
    """
    return isamAppliance.invoke_get("Retrieve configuration of SCIM ISAM user settings",
                                    "{0}/urn:ietf:params:scim:schemas:extension:isam:1.0:User".format(uri),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)

def set(isamAppliance, ldap_connection=None, isam_domain=None, update_native_users=None, connection_type=None, attrs_dir=None, enforce_password_policy=None, check_mode=False, force=False):
    """
    Updating the ISAM user SCIM configuration settings
    """
    ret_obj = get(isamAppliance)
    ret_obj = ret_obj['data']['urn:ietf:params:scim:schemas:extension:isam:1.0:User']

    new_obj = {}

    update_required = False

    if ldap_connection is None:
        if 'ldap_connection' in ret_obj:
            new_obj['ldap_connection'] = ret_obj['ldap_connection']
    else:
        new_obj['ldap_connection'] = ldap_connection
        if 'ldap_connection' in ret_obj:
            if ret_obj['ldap_connection'] != ldap_connection:
                update_required = True
        else:
            update_required = True

    if isam_domain is None:
        if 'isam_domain' in ret_obj:
            new_obj['isam_domain'] = ret_obj['isam_domain']
    else:
        new_obj['isam_domain'] = isam_domain
        if 'isam_domain' in ret_obj:
            if ret_obj['isam_domain'] != isam_domain:
                update_required = True
        else:
            update_required = True

    if update_native_users is None:
        if 'update_native_users' in ret_obj:
            new_obj['update_native_users'] = ret_obj['update_native_users']
    else:
        new_obj['update_native_users'] = update_native_users
        if 'update_native_users' in ret_obj:
            if ret_obj['update_native_users'] != update_native_users:
                update_required = True
        else:
            update_required = True

    if connection_type is None:
        if 'connection_type' in ret_obj:
            new_obj['connection_type'] = ret_obj['connection_type']
    else:
        new_obj['connection_type'] = connection_type
        if 'connection_type' in ret_obj:
            if ret_obj['connection_type'] != connection_type:
                update_required = True
        else:
            update_required = True

    if attrs_dir is None:
        if 'attrs_dir' in ret_obj:
            new_obj['attrs_dir'] = ret_obj['attrs_dir']
    else:
        new_obj['attrs_dir'] = attrs_dir
        if 'attrs_dir' in ret_obj:
            if ret_obj['attrs_dir'] != attrs_dir:
                update_required = True
        else:
            update_required = True

    if enforce_password_policy is None:
        if 'enforce_password_policy' in ret_obj:
            new_obj['enforce_password_policy'] = ret_obj['enforce_password_policy']
    else:
        new_obj['enforce_password_policy'] = enforce_password_policy
        if 'enforce_password_policy' in ret_obj:
            if ret_obj['enforce_password_policy'] != enforce_password_policy:
                update_required = True
        else:
            update_required = True

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Updating the ISAM user SCIM configuration settings",
                "/mga/scim/configuration/urn:ietf:params:scim:schemas:extension:isam:1.0:User",
                new_obj, requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(changed=False)


import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/policyattachments"


def get_all(isamAppliance, filter=None, sortBy=None, check_mode=False, force=False):
    """
    Retrieve a list of configured resources
    """
    return isamAppliance.invoke_get(
        "Retrieve a list of configured resources",
        "{0}{1}".format(uri, tools.create_query_string(filter=filter, sortBy=sortBy)),
    )


def get(isamAppliance, server, resourceUri, check_mode=False, force=False):
    """
    Retrieve a specific configured resource
    """
    ret_obj = search(
        isamAppliance,
        server=server,
        resourceUri=resourceUri,
        check_mode=check_mode,
        force=force,
    )
    resource_id = ret_obj["data"]

    if resource_id == {}:
        logger.info(
            "Resource {0}/{1} had no match, skipping retrieval.".format(
                server, resourceUri
            )
        )
        return isamAppliance.create_return_object()
    else:
        return isamAppliance.invoke_get(
            "Retrieve a specific configured resource",
            "{0}/{1}".format(uri, resource_id),
        )


def get_attachments(isamAppliance, server, resourceUri, check_mode=False, force=False):
    """
    Retrieve a list of attachments for a resource
    """
    ret_obj = search(
        isamAppliance,
        server=server,
        resourceUri=resourceUri,
        check_mode=check_mode,
        force=force,
    )
    resource_id = ret_obj["data"]

    if resource_id == {}:
        logger.info(
            "Resource {0}/{1} had no match, skipping retrieval.".format(
                server, resourceUri
            )
        )
        return isamAppliance.create_return_object()
    else:
        return isamAppliance.invoke_get(
            "Retrieve a list of attachments for a resource",
            "{0}/{1}/policies".format(uri, resource_id),
        )


def search(isamAppliance, server, resourceUri, force=False, check_mode=False):
    """
    Search server/resource uri id by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj["data"]:
        if obj["resourceUri"] == resourceUri and obj["server"] == server:
            logger.info(
                "Found server/resourceUri {0}/{1} id: {2}".format(
                    server, resourceUri, obj["id"]
                )
            )
            return_obj["data"] = obj["id"]
            return_obj["rc"] = 0

    return return_obj


def authenticate(
    isamAppliance, username, password, domain="Default", check_mode=False, force=False
):
    """
    Authenticate with Security Access Manager
    """
    ret_obj = isamAppliance.invoke_post(
        "Authenticate with Security Access Manager",
        "{0}/pdadmin".format(uri),
        {
            "command": "setCredential",
            "username": username,
            "password": password,
            "domain": domain,
        },
    )

    ret_obj["changed"] = (
        False  # Authentication call changes nothing on the appliance (HTTP POST defaults to change)
    )

    return ret_obj


def get_resources(isamAppliance, object="/WebSEAL", check_mode=False, force=False):
    """
    Retrieve a list of resources in a protected object space
    """
    ret_obj = isamAppliance.invoke_post(
        "Retrieve a list of resources in a protected object space",
        "{0}/pdadmin".format(uri),
        {"command": "object list", "object": object},
    )

    ret_obj["changed"] = (
        False  # Resource get call changes nothing on the appliance (HTTP POST defaults to change)
    )

    return ret_obj


def config(
    isamAppliance,
    server,
    resourceUri,
    policies=[],
    policyType=None,
    policyCombiningAlgorithm=None,
    cache=None,
    check_mode=False,
    force=False,
):
    """
    Configure a resource
    Note: Please input policies with policy names (it will be converted to id's), like so:
     [{'name': '<policy name>', 'type': 'policy'}, {'name': '<policyset name>', 'type': 'policyset'},
      {'name': '<definition name>', 'type': 'definition'}]
    """
    warnings = []
    if force is False:
        ret_obj = search(isamAppliance, server, resourceUri)

    if force is True or ret_obj["data"] == {}:
        json_data = {"server": server, "resourceUri": resourceUri}
        if policyType is not None:
            if tools.version_compare(isamAppliance.facts["version"], "9.0.6.0") < 0:
                warnings.append(
                    "Appliance at version: {0}, policyType: {1} is not supported. Needs 9.0.6.0 or higher. Ignoring policyType for this call.".format(
                        isamAppliance.facts["version"], cache
                    )
                )
            else:
                json_data["type"] = policyType

        json_data["policies"] = _convert_policy_name_to_id(isamAppliance, policies)
        if policyCombiningAlgorithm is not None:
            json_data["policyCombiningAlgorithm"] = policyCombiningAlgorithm
        if cache is not None:
            if tools.version_compare(isamAppliance.facts["version"], "9.0.3.0") < 0:
                warnings.append(
                    "Appliance at version: {0}, cache: {1} is not supported. Needs 9.0.3.0 or higher. Ignoring cache for this call.".format(
                        isamAppliance.facts["version"], cache
                    )
                )
            else:
                json_data["cache"] = int(cache)
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post(
                "Configure a resource", uri, json_data, warnings=warnings
            )

    return isamAppliance.create_return_object()


def update(
    isamAppliance,
    server,
    resourceUri,
    policyCombiningAlgorithm,
    cache=None,
    check_mode=False,
    force=False,
):
    """
    Update the policy attachment combining algorithm
    """
    warnings = []
    ret_obj = get(isamAppliance, server, resourceUri)

    if force or (
        ret_obj["data"] != {} and (
            ret_obj["data"]["policyCombiningAlgorithm"] != policyCombiningAlgorithm or ret_obj["data"]["cache"] != cache
        )
    ):
        json_data = {"policyCombiningAlgorithm": policyCombiningAlgorithm}
        if cache is not None:
            if tools.version_compare(isamAppliance.facts["version"], "9.0.3.0") < 0:
                warnings.append(
                    "Appliance at version: {0}, cache: {1} is not supported. Needs 9.0.3.0 or higher. Ignoring cache for this call.".format(
                        isamAppliance.facts["version"], cache
                    )
                )
            else:
                json_data["cache"] = int(cache)
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Update the policy attachment combining algorithm and cache",
                "{0}/{1}/properties".format(uri, ret_obj["data"]["id"]),
                json_data,
                warnings=warnings,
            )

    return isamAppliance.create_return_object()


def update_attachments(
    isamAppliance,
    server,
    resourceUri,
    attachments,
    action,
    check_mode=False,
    force=False,
):
    """
    Update the attachments for a resource

    Provide attachemnts like so:
     [{'name': '<policy name>', 'type': 'policy'}, {'name': '<policyset name>', 'type': 'policyset'},
      {'name': '<definition name>', 'type': 'definition'}]
    """
    ret_obj = get(isamAppliance, server, resourceUri)
    cur_policies = ret_obj["data"]["policies"]

    if force is True or _check(cur_policies, attachments, action) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            new_policies = _convert_policy_name_to_id(isamAppliance, attachments)
            return isamAppliance.invoke_put(
                "Update the attachments for a resource",
                "{0}/{1}/policies{2}".format(
                    uri, ret_obj["data"]["id"], tools.create_query_string(action=action)
                ),
                new_policies,
            )
    return isamAppliance.create_return_object()


def _check(policies, attachments, action):
    # Check and see if there is even one match
    full_match = True
    partial_match = False
    for new_pol in attachments:
        match = False
        for cur_pol in policies:
            if (
                new_pol["name"] == cur_pol["name"] and new_pol["type"] == cur_pol["type"]
            ):
                logger.info(
                    "At least one policy already exists {0}/{1}".format(
                        new_pol["type"], new_pol["name"]
                    )
                )
                match = True
                break
        if match is False:
            full_match = False
        elif partial_match is False:
            partial_match = True

    # Add will be rejected if there is even one match
    if action == "add":
        logger.info(
            "Check if there is at least one match returned: {0}".format(partial_match)
        )
        return not partial_match
    # Delete will be rejected if there is a partial match (has to be full match)
    elif action == "remove":
        logger.info(
            "Check if there is a full match of provided policies: {0}".format(
                full_match
            )
        )
        return full_match
    # Set requires that there be a full match and the number of elements match
    elif action == "set":
        if len(policies) == len(attachments) and full_match is True:
            return False
        else:
            return True
    # Force delete will be rejected if there is no match
    elif action == "force_remove":
        logger.info(
            "Check if there is at least one match returned: {0}".format(partial_match)
        )
        return partial_match
    else:
        from ibmsecurity.appliance.ibmappliance import IBMError

        raise IBMError("999", "Unknown action provided: {0}".format(action))


def publish(isamAppliance, server, resourceUri, check_mode=False, force=False):
    """
    Publish the policy attachments for a resource
    """
    ret_obj = get(isamAppliance, server, resourceUri)

    if force or (
        ret_obj["data"] != {} and (
            ret_obj["data"]["deployrequired"] or not ret_obj["data"]["deployed"]
        ) and ret_obj["data"]["policies"] != []
    ):
        if check_mode:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Publish the policy attachments for a resource",
                "{0}/deployment/{1}".format(uri, ret_obj["data"]["id"]),
                {},
            )

    return isamAppliance.create_return_object()


def publish_list(isamAppliance, attachments, check_mode=False, force=False):
    """
    Publish a list of policy attachments

    Note: provide attachments like so:
    [{'server': '<server1>', 'resourceUri': '<resourceuri1>'}, {'server': '<server2>', 'resourceUri': '<resourceuri2>'}]
    """
    id_list = []
    for attach in attachments:
        ret_obj = get(isamAppliance, attach["server"], attach["resourceUri"])
        if force is True or ret_obj["data"]["deployrequired"] is True:
            id_list.append(ret_obj["data"]["id"])
    logger.debug("Attachments: {0}".format(id_list))

    if len(id_list) > 0:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Publish a list of policy attachments",
                "{0}/deployment".format(uri),
                {"policyAttachmentIds": ",".join(id_list)},
            )

    return isamAppliance.create_return_object()


def _convert_policy_name_to_id(isamAppliance, policies):
    """
    Converts this:
    [{'name': '<policy name>', 'type': 'policy'}, {'name': '<policyset name>', 'type': 'policyset'},
     {'name': '<definition name>}, 'type': 'definition'}]
    to:
    [{'id': '<policy id>', 'type': 'policy'}, {'id': '<policyset id>', 'type': 'policyset'},
     {'id': '<definition id>, 'type': 'definition'}]
    """
    pol_ids = []
    import ibmsecurity.isam.aac.access_control.policies
    import ibmsecurity.isam.aac.access_control.policy_sets
    import ibmsecurity.isam.aac.api_protection.definitions

    for pol in policies:
        name = pol["name"]
        type = pol["type"]
        if type == "policy":
            ret_obj = ibmsecurity.isam.aac.access_control.policies.search(
                isamAppliance, name
            )
            pol_id = ret_obj["data"]
            if pol_id != {}:
                logger.debug("Converting policy {0} to ID: {1}".format(name, pol_id))
            else:
                logger.warning("Unable to find policy {0}, skipping.".format(name))
        elif type == "policyset":
            ret_obj = ibmsecurity.isam.aac.access_control.policy_sets.search(
                isamAppliance, name
            )
            pol_id = ret_obj["data"]
            if pol_id != {}:
                logger.debug(
                    "Converting policy set {0} to ID: {1}".format(name, pol_id)
                )
            else:
                logger.warning("Unable to find policy set {0}, skipping.".format(name))
        elif type == "definition":
            ret_obj = ibmsecurity.isam.aac.api_protection.definitions.search(
                isamAppliance, name
            )
            pol_id = ret_obj["data"]
            if pol_id != {}:
                logger.debug(
                    "Converting api definition {0} to ID: {1}".format(name, pol_id)
                )
            else:
                logger.warning(
                    "Unable to find api definition {0}, skipping.".format(name)
                )
        else:
            from ibmsecurity.appliance.ibmappliance import IBMError

            raise IBMError(
                "999", "Policy specified with unknown type: {0}/{1}".format(type, name)
            )

        pol_ids.append({"id": pol_id, "type": type})

    return pol_ids


def _convert_policy_id_to_name(isamAppliance, policies):
    """
    Converts this:
    [{'id': '<policy id>', 'type': 'policy'}, {'id': '<policyset id>', 'type': 'policyset'},
     {'id': '<definition id>, 'type': 'definition'}]
    to:
    [{'name': '<policy name>', 'type': 'policy'}, {'name': '<policyset name>', 'type': 'policyset'},
     {'name': '<definition name>}, 'type': 'definition'}]
    """
    pol_ids = []
    import ibmsecurity.isam.aac.access_control.policies
    import ibmsecurity.isam.aac.access_control.policy_sets
    import ibmsecurity.isam.aac.api_protection.definitions

    for pol in policies:
        pol_id = pol["id"]
        type = pol["type"]
        if type == "policy":
            ret_obj = ibmsecurity.isam.aac.access_control.policies._get(
                isamAppliance, pol_id
            )
            pol_name = ret_obj["data"]["name"]
            if pol_name != {}:
                logger.debug(
                    "Converting policy {0} to Name: {1}".format(pol_id, pol_name)
                )
            else:
                logger.warning("Unable to find policy {0}, skipping.".format(pol_id))
        elif type == "policyset":
            ret_obj = ibmsecurity.isam.aac.access_control.policy_sets._get(
                isamAppliance, pol_id
            )
            pol_name = ret_obj["data"]["name"]
            if pol_name != {}:
                logger.debug(
                    "Converting policy set {0} to Name: {1}".format(pol_id, pol_name)
                )
            else:
                logger.warning(
                    "Unable to find policy set {0}, skipping.".format(pol_id)
                )
        elif type == "definition":
            ret_obj = ibmsecurity.isam.aac.api_protection.definitions._get(
                isamAppliance, pol_id
            )
            pol_name = ret_obj["data"]["name"]
            if pol_name != {}:
                logger.debug(
                    "Converting api definition {0} to Name: {1}".format(
                        pol_id, pol_name
                    )
                )
            else:
                logger.warning(
                    "Unable to find api definition {0}, skipping.".format(pol_id)
                )
        else:
            from ibmsecurity.appliance.ibmappliance import IBMError

            raise IBMError(
                "999",
                "Policy specified with unknown type: {0}/{1}".format(type, pol_id),
            )

        pol_ids.append({"name": pol_name, "type": type})

    return pol_ids


def delete(isamAppliance, server, resourceUri, check_mode=False, force=False):
    """
    Delete a configured resource
    """
    ret_obj = search(
        isamAppliance, server, resourceUri, check_mode=check_mode, force=force
    )
    resourceID = ret_obj["data"]

    if resourceID == {}:
        logger.info(
            "ResourceURI {0}/{1} not found, skipping delete.".format(
                server, resourceUri
            )
        )
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a configured resource", "{0}/{1}".format(uri, resourceID)
            )

    return isamAppliance.create_return_object()


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Policy Sets between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1["data"]:
        del obj["id"]
        del obj["userlastmodified"]
        del obj["lastmodified"]
        del obj["datecreated"]
        del obj["lastdeployed"]
        for pol in obj["policies"]:
            del pol["id"]
            del pol["lastmodified"]
    for obj in ret_obj2["data"]:
        del obj["id"]
        del obj["userlastmodified"]
        del obj["lastmodified"]
        del obj["datecreated"]
        del obj["lastdeployed"]
        for pol in obj["policies"]:
            del pol["id"]
            del pol["lastmodified"]

    return tools.json_compare(
        ret_obj1,
        ret_obj2,
        deleted_keys=[
            "id",
            "userlastmodified",
            "lastmodified",
            "datecreated",
            "lastdeployed",
        ],
    )



import logging

from ibmsecurity.utilities import tools

try:
    basestring
except NameError:
    basestring = (str, bytes)

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/audit/components"
requires_modules = ["mga", "federation"]
requires_version = None


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve audit configuration
    """
    return isamAppliance.invoke_get("Retrieve all audit component configuration", uri, requires_modules=requires_modules,
                                    requires_version=requires_version)


def search(isamAppliance, component_name: str, check_mode=False, force=False):
    """
    Get the id for the component by (group) name
    """
    ret_obj = None

    ret_obj = get_all(isamAppliance)
    if ret_obj.get("data", None):
        for obj in ret_obj.get("data"):
            if obj['group'] == component_name:
                logger.info(f"Found name {component_name} id: {obj['id']}")
                return obj['id']
        return None
    else:
        return None


def get(isamAppliance, group_name: str = None, component_id: str = None, type_id: str = None, check_mode=False, force=False):
    """
    Retrieve specific audit configuration component group by component_id, by type or by group name
    """
    requires_version = None
    warnings = []
    if component_id is None and type_id is None and group_name is None:
        warnings = ['No group_name, component_id nor type_id passed']
        return isamAppliance.create_return_object(warnings)
    elif group_name:
        # translate name to component_id
        component_id = search(isamAppliance, component_name=group_name)
        if component_id:
            uri_part = component_id
        else:
            warnings = [f"Cannot find group by name of: {group_name}"]
            return isamAppliance.create_return_object(warnings)
    elif component_id:
        # Ignore type_id in this case
        uri_part = component_id
    else:
        # valid values are runtime or management
        if type_id in ['runtime', 'management']:
            requires_version = "10.0.7.0"
            uri_part = type_id
        else:
            warnings = [f"Invalid type_id passed to function: {type_id}"]
            return isamAppliance.create_return_object(warnings)
    return isamAppliance.invoke_get("Retrieve audit configuration component ", f"{uri}/{uri_part}", requires_modules=requires_modules,
                                    requires_version=requires_version, warnings=warnings)


def set(isamAppliance, component_id: str, enabled=True, check_mode=False, force=False):
    """
    Update Audit Configuration Component by id
    This simply enables or disables the group.
    TODO: Add set by type and set_all
    """
    if isinstance(enabled, str):
        if enabled.upper() in ['TRUE', 'YES']:
            enabled = True
        else:
            enabled = False
    update_required = _check(isamAppliance, component_id, enabled)
    if enabled:
        json_data = {
            'enabled': True
        }
    else:
        json_data = {
            'enabled': False
        }
    if force or update_required:
        if check_mode:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update Audit Configuration Component",
                f"{uri}/{component_id}",
                json_data,
                requires_modules=requires_modules,
                requires_version=requires_version)


def _check(isamAppliance, component_id: str, enabled: bool):
    """
    Check and return True if update needed
    """
    update_required = False

    ret_obj = get(isamAppliance, component_id=component_id)
    cmp_cfg = ret_obj.get("data", None)
    logger.debug(f"\n\n{cmp_cfg}\n\n")
    if cmp_cfg is not None and str(cmp_cfg.get('enabled', "frottekop")) != str(enabled):
        update_required = True
        logger.debug(f"\n\nAudit Configuration Component requires an update {cmp_cfg.get('enabled')} <> {enabled}")
    else:
        logger.warning("Audit Configuration Component does not need an update or does not exist.")

    return update_required


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Audit Configuration Components between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['id']
    for obj in ret_obj2['data']:
        del obj['id']

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id'])



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)

uri = "/isam/authzserver"
requires_modules = None
requires_version = None
version = "v1"
requires_model = "Appliance"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve all authorization servers
    """
    return isamAppliance.invoke_get(description="Retrieving all authorization servers",
                                    uri="{0}/{1}".format(uri, version),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version,
                                    requires_model=requires_model)


def _check(isamAppliance, id):
    """
    Check if authorization server is already created or not

    :param isamAppliance:
    :return:
    """
    ret_obj = get(isamAppliance)
    azsrv_exists, warnings = False, ret_obj['warnings']

    if not warnings:
        logger.debug("Looking for existing authorization servers in: {0}".format(ret_obj['data']))
        if ret_obj['data']:
            for acld in ret_obj['data']:
                if acld['id'] == id:
                    logger.debug("Found authorization server: {0}".format(id))
                    azsrv_exists = True

    return azsrv_exists, warnings


def add(isamAppliance, inst_name, admin_pwd, addresses, hostname='localhost', authport='7136', adminport='7137',
        domain='Default',
        admin_id='sec_master', ssl='no', ssl_port=None, keyfile=None, keyfile_label=None, check_mode=False,
        force=False):
    """
    Add an authorization server

    :param isamAppliance, inst_name, admin_pwd, addresses:
    :return:
    """
    azsrv_exists, warnings = _check(isamAppliance, inst_name)

    if warnings:
        return isamAppliance.create_return_object(warnings=warnings)

    if force is True or azsrv_exists is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post(description="Add an authorization server",
                                             uri="{0}/{1}".format(uri, version),
                                             data={
                                                 "inst_name": inst_name,
                                                 "hostname": hostname,
                                                 "authport": authport,
                                                 "adminport": adminport,
                                                 "domain": domain,
                                                 "admin_id": admin_id,
                                                 "admin_pwd": admin_pwd,
                                                 "addresses": addresses,
                                                 "ssl": ssl,
                                                 "ssl_port": ssl_port,
                                                 "keyfile": keyfile,
                                                 "keyfile_label": keyfile_label
                                             },
                                             requires_modules=requires_modules,
                                             requires_version=requires_version,
                                             requires_model=requires_model)

    return isamAppliance.create_return_object(warnings="Authorization server {0} already exists".format(inst_name))


def delete(isamAppliance, id, admin_pwd, admin_id='sec_master', check_mode=False, force=False):
    """
    Unconfigure existing runtime component
    """

    azsrv_exists, warnings = _check(isamAppliance, id)

    if warnings:
        return isamAppliance.create_return_object(warnings=warnings)

    if force is True or force == "yes" or azsrv_exists is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            force_yn = "no"
            if force is True or force == "yes":
                force_yn = "yes"
            return isamAppliance.invoke_put(description="Remove a authorization server",
                                            uri="{0}/{1}/{2}".format(uri, id, version),
                                            data={
                                                "operation": "unconfigure",
                                                "force": force_yn,
                                                "admin_id": admin_id,
                                                "admin_pwd": admin_pwd
                                            },
                                            requires_modules=requires_modules,
                                            requires_version=requires_version,
                                            requires_model=requires_model)

    return isamAppliance.create_return_object(
        warnings="The authorization server instance specified in the request does not exist. Check that the authorization server instance is correct: {0}".format(
            id))


def execute(isamAppliance, id, operation="restart", admin_id="sec_master", admin_pwd=None, check_mode=False,
            force=False, warnings=[]):
    """
    Execute an operation on runtime component

    :param isamAppliance:
    :param operation:
    :return:
    """
    ret_obj = get(isamAppliance)
    warnings = ret_obj['warnings']
    if warnings and 'Docker' in warnings[0]:
        return isamAppliance.create_return_object(warnings=ret_obj['warnings'])

    for acld in ret_obj['data']:
        if acld['id'] == id:
            if (force is True or
                    (acld['restart'] == "true" and operation == "restart") or
                    (acld['started'] == 'yes' and operation == "stop") or
                    (acld['started'] == 'no' and operation == "start") or
                    (operation == "renew")):
                if check_mode is True:
                    return isamAppliance.create_return_object(changed=True)
                else:
                    body_json = {
                        "operation": operation
                    }
                    if operation == "renew":
                        if admin_pwd is not None:
                            body_json["admin_id"] = admin_id
                            body_json["admin_pwd"] = admin_pwd
                        else:
                            warnings.append("admin_pwd must be specified when operation is renew")
                            return isamAppliance.create_return_object(warnings=warnings)
                    return isamAppliance.invoke_put(description="Execute an operation on authorization server",
                                                    uri="{0}/{1}/{2}".format(uri, id, version),
                                                    data=body_json,
                                                    requires_modules=requires_modules,
                                                    requires_version=requires_version, warnings=warnings)
            break
    if _check(isamAppliance, id) is False:
        return isamAppliance.create_return_object(
            warnings="The authorization server instance specified in the request does not exist. Check that the authorization server instance is correct: {0}".format(
                id))
    else:
        return isamAppliance.create_return_object()


def compare(isamAppliance1, isamAppliance2):
    """
    Compare list of authorization servers between 2 appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    for acld in ret_obj1['data']:
        del acld['started']
        del acld['restart']
        del acld['version']
    for acld in ret_obj2['data']:
        del acld['started']
        del acld['restart']
        del acld['version']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['restart', 'started', 'version'])



import logging
import ibmsecurity.utilities.tools
import os.path
from ibmsecurity.utilities.tools import files_same, get_random_temp_dir
import shutil

logger = logging.getLogger(__name__)


def get_all(isamAppliance, instance_id, check_mode=False, force=False):
    """
    Retrieving the current administration pages root contents
    """
    return isamAppliance.invoke_get("Retrieving the current administration pages root contents",
                                    "/wga/reverseproxy/{0}/management_root?recursive=yes".format(instance_id))


def get(isamAppliance, instance_id, id, check_mode=False, force=False):
    """
    Retrieving the contents of a file in the administration pages root
    """
    return isamAppliance.invoke_get("Retrieving the contents of a file in the administration pages root",
                                    "/wga/reverseproxy/{0}/management_root/{1}".format(instance_id, id))


def _check(isamAppliance, instance_id, id, name):
    ret_obj = get_all(isamAppliance, instance_id)

    file_name = os.path.join(id, name)

    return _parse_id(ret_obj['data'], file_name)


def _check_file(isamAppliance, instance_id, id):
    """
    Check whether file exists
    :param isamAppliance:
    :param instance_id:
    :param id:
    :return: True/False
    """
    ret_obj = get(isamAppliance, instance_id, id)

    logger.info(ret_obj['data'])

    if ret_obj['rc'] == 0:
        return True
    else:
        return False


def _parse_id(contents, file_name):
    """
    Recursively parse and find the id for a given file name

    :param contents:
    :param file_name:
    :return id:
    """
    try:
        split_file = file_name.split('/', 1)
        cur_file = split_file[0]
        rest_file = split_file[1]
    except:
        rest_file = ''
    for file in contents:
        if file['name'] == cur_file:
            if rest_file == '':
                if file['type'] == 'File':
                    return file['id']
                else:
                    return None
            else:
                if len(file['children']) == 0:
                    return None
                else:
                    return _parse_id(file['children'], rest_file)

    return None


def create(isamAppliance, instance_id, id, name, contents=None, check_mode=False, force=False):
    """
    Creating a file in the administration pages root

    :param isamAppliance:
    :param instance_id:
    :param id:
    :param name:
    :param contents:
    :param check_mode:
    :param force:
    :return:
    """
    if force is True or _check(isamAppliance, instance_id, id, name) == None:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Creating a file in the administration pages root",
                "/wga/reverseproxy/{0}/management_root/{1}".format(instance_id, id),
                {
                    'file_name': name,
                    'type': 'file',
                    'contents': contents
                })

    return isamAppliance.create_return_object()


def update(isamAppliance, instance_id, id, filename=None, contents=None, check_mode=False, force=False):
    """
    Update a file in the administration pages root
    :param isamAppliance:
    :param instance_id:
    :param id:
    :param name:
    :param contents:
    :param check_mode:
    :param force:
    :return:
    """

    if force is True or _check_file(isamAppliance, instance_id, id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            if filename is not None:
                return isamAppliance.invoke_put_files(
                    "Update a file in the administration page root",
                    "/wga/reverseproxy/{0}/management_root/{1}".format(instance_id, id),
                    [
                        {
                            'file_formfield': 'file',
                            'filename': filename,
                            'mimetype': 'application/octet-stream'
                        }
                    ],
                    {
                        'file': filename,
                        'type': 'file'
                    })
            elif contents is not None:
                return isamAppliance.invoke_put_files(
                    "Update a file in the administration page root",
                    "/wga/reverseproxy/{0}/management_root/{1}".format(instance_id, id),
                    {
                        'contents': contents,
                        'type': 'file'
                    })
            else:
                return isamAppliance.create_return_object(
                    warnings=["Either contents or filename parameter need to be provided. Skipping update request."])


def delete(isamAppliance, instance_id, id, check_mode=False, force=False):
    """
    Deleting a file in the administration pages root

    :param isamAppliance:
    :param instance_id:
    :param id:
    :param check_mode:
    :param force:
    :return:
    """
    if force is True or _check(isamAppliance, instance_id, id, '') != None:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Deleting a file in the administration pages root",
                "/wga/reverseproxy/{0}/management_root/{1}".format(instance_id, id))

    return isamAppliance.create_return_object()


def rename(isamAppliance, instance_id, id, new_name, check_mode=False, force=False):
    """
    Deleting a file in the administration pages root

    :param isamAppliance:
    :param instance_id:
    :param id:
    :param new_name:
    :param check_mode:
    :param force:
    :return:
    """
    file_id = None
    if force is False:
        file_id = _check(isamAppliance, instance_id, id, '')

    if force is True or file_id != None:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Renaming a file in the administration pages root",
                "/wga/reverseproxy/{0}/management_root/{1}".format(instance_id, id),
                {
                    'id': file_id,
                    'new_name': new_name,
                    'type': 'file'
                })

    return isamAppliance.create_return_object()


def export_file(isamAppliance, instance_id, id, filename, check_mode=False, force=False):
    """
    Exporting a file in the administration pages root

    :param isamAppliance:
    :param instance_id:
    :param id:
    :param filename:
    :param check_mode:
    :param force:
    :return:
    """
    if force is True or (_check(isamAppliance, instance_id, id, name='') is not None):
        if check_mode is False:
            return isamAppliance.invoke_get_file(
                "Exporting a file in the administration pages root",
                "/wga/reverseproxy/{0}/management_root/{1}?export=true".format(instance_id, id), filename)

    return isamAppliance.create_return_object()


def _check_import(isamAppliance, instance_id, id, filename, check_mode=False):
    """
    Checks if file on the Appliance (id) exists and if so, whether it is different from filename
    :param isamAppliance:
    :param instance_id:
    :param id:
    :param filename:
    :return:
    """
    tmpdir = get_random_temp_dir()
    tmp_original_file = os.path.join(tmpdir, os.path.basename(id))
    if (_check(isamAppliance, instance_id, id, '') is not None):
        export_file(isamAppliance, instance_id, id, tmp_original_file, check_mode=False, force=True)
        logger.debug("file already exists on appliance")
        if files_same(tmp_original_file, filename):
            logger.debug("files are the same, so we don't want to do anything")
            shutil.rmtree(tmpdir)
            return False
        else:
            logger.debug("files are different, so we delete existing file in preparation for import")
            delete(isamAppliance, instance_id, id, check_mode=check_mode, force=True)
            shutil.rmtree(tmpdir)
            return True
    else:
        logger.debug("file does not exist on appliance, so we'll want to import")
        shutil.rmtree(tmpdir)
        return True


def import_file(isamAppliance, instance_id, id, filename, check_mode=False, force=False):
    """
    Importing a file in the administration pages root.
    Note that _check_import() checks that the file (id) is present and if so, whether it is
    the same and the one being imported.  If it's the same, there is no need to import the new file.
    """
    if force is True or _check_import(isamAppliance, instance_id, id, filename, check_mode=check_mode):
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post_files(
                "Importing a file in the administration pages root",
                "/wga/reverseproxy/{0}/management_root/{1}".format(instance_id, id),
                [
                    {
                        'file_formfield': 'file',
                        'filename': filename,
                        'mimetype': 'application/octet-stream'
                    }
                ],
                {
                    'type': 'file',
                    'force': force
                })

    return isamAppliance.create_return_object()


def compare(isamAppliance1, isamAppliance2, instance_id):
    ret_obj1 = get_all(isamAppliance1, instance_id)
    ret_obj2 = get_all(isamAppliance2, instance_id)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
from ibmsecurity.utilities.tools import json_sort

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/alias_settings"
requires_modules = ["federation"]
requires_version = "9.0.6.0"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve alias associations
    """
    return isamAppliance.invoke_get("Retrieve alias associations", "{0}".format(uri),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def update(isamAppliance, aliasDBType, properties=None, check_mode=False, force=False):
    """
    Update the alias settings
    """
    ret_obj = get(isamAppliance)
    update_required = False

    if ret_obj['data']['aliasDBType'] == aliasDBType:
        if 'properties' in ret_obj['data']:
            props = ret_obj['data']['properties']
            sorted_props = json_sort(props)
            sorted_properties = json_sort(properties)
            if sorted_props != sorted_properties:
                update_required = True
    else:
        update_required = True

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update the alias settings",
                "{0}".format(uri),
                {
                    'aliasDBType': aliasDBType,
                    'properties': properties
                },
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()



import logging
import ibmsecurity.utilities.tools
from ibmsecurity.isam.web.reverse_proxy import junctions

logger = logging.getLogger(__name__)
requires_modules = ["wga"]
requires_version = "9.0.4.0"


def config(isamAppliance, instance_id, hostname='127.0.0.1', port=443, username=None, password=None,
           junction="/mga", reuse_certs=False, reuse_acls=False, api=False, browser=False, auth_register=None, fapi_compliant=None,
           check_mode=False, force=False):
    """
    Oauth and Oidc configuration for a reverse proxy instance

    :param isamAppliance:
    :param instance_id:
    :param junction:
    :param hostname:
    :param port:
    :param username:
    :param password:
    :param reuse_certs:
    :param reuse_acls:
    :param api:
    :param browser:
    :param auth_register:
    :param check_mode:
    :param force:
    :param fapi_compliant:
    :return:
    """
    if username is None:
        logger.info("Required parameter username missing. Skipping config.")
        return isamAppliance.create_return_object(warning=["Required parameter username missing. Skipping config."])

    if password is None:
        logger.info("Required parameter password missing. Skipping config.")
        return isamAppliance.create_return_object(warning=["Required parameter password missing. Skipping config."])

    warnings = [
        "Idempotency logic will check for existence of {} junction. Use force=True to override.".format(junction)]
    if force is True or _check_config(isamAppliance, instance_id, junction) is False:
        json_data = {
            "junction": junction,
            "hostname": hostname,
            "port": port,
            "username": username,
            "password": password,
            "api": api,
            "browser": browser,
            "reuse_certs": reuse_certs,
            "reuse_acls": reuse_acls
        }
        if auth_register is not None:
            if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts["version"], "9.0.5.0") < 0:
                warnings.append(
                    "Appliance at version: {0}, auth_register: {1} is not supported. Needs 9.0.5.0 or higher. Ignoring auth_register for this call.".format(
                        isamAppliance.facts["version"], auth_register))
            else:
                json_data["auth_register"] = auth_register
        if fapi_compliant is not None:
            if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts["version"], "10.0.0.0") < 0:
                warnings.append(
                    "Appliance at version: {0}, fapi_compliant: {1} is not supported. Needs 10.0.0.0 or higher. Ignoring fapi_compliant for this call.".format(
                        isamAppliance.facts["version"], fapi_compliant))
            else:
                json_data["fapi_compliant"] = fapi_compliant 
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post(
                "OAuth configuration for a reverse proxy instance",
                "/wga/reverseproxy/{0}/oauth_config".format(instance_id), json_data, warnings=warnings,
                requires_modules=requires_modules, requires_version=requires_version)
    return isamAppliance.create_return_object(warnings=warnings)

def _check_config(isamAppliance, instance_id, junction):
    """
    Check if the junction for oauth already created

    :param isamAppliance:
    :param instance_id:
    :param junction:
    :return:
    """
    ret_obj = junctions.get_all(isamAppliance, instance_id)

    for j in ret_obj['data']:
        if j['id'] == junction:
            logger.info("Junction {} was found - hence oauth config must have already executed.".format(junction))
            return True

    return False



import logging
import ibmsecurity.utilities.tools

try:
    basestring
except NameError:
    basestring = (str, bytes)

logger = logging.getLogger(__name__)

# URI for this module
uri = "/core/dca_updates_cfg"
requires_modules = None
requires_version = None


def get(isamAppliance, check_mode=False, force=False):
    """
    Get Application Database Settings
    """
    return isamAppliance.invoke_get("Get Application Database Settings", uri, requires_modules=requires_modules,
                                    requires_version=requires_version)


def set(isamAppliance, enableIprAutoUpdate=True, useProxy=False, proxyHost=None, proxyPort=None, useProxyAuth=False,
        proxyUser=None, proxyPwd=None, enableAutoUpdate=True, enableIprFeedback=False, enableWeblearn=False,
        includeIprInfo=False, check_mode=False, force=False):
    """
    Set Application Database Settings
    """
    update_required, json_data = _check(isamAppliance, enableIprAutoUpdate=enableIprAutoUpdate, useProxy=useProxy,
                                        proxyHost=proxyHost, proxyPort=proxyPort, useProxyAuth=useProxyAuth,
                                        proxyUser=proxyUser, proxyPwd=proxyPwd, enableAutoUpdate=enableAutoUpdate,
                                        enableIprFeedback=enableIprFeedback, enableWeblearn=enableWeblearn,
                                        includeIprInfo=includeIprInfo)

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put("Set Application Database Settings", uri, json_data,
                                            requires_modules=requires_modules,
                                            requires_version=requires_version)

    return isamAppliance.create_return_object()


def _check(isamAppliance, enableIprAutoUpdate, useProxy, proxyHost, proxyPort, useProxyAuth, proxyUser, proxyPwd,
           enableAutoUpdate, enableIprFeedback, enableWeblearn, includeIprInfo):
    update_required = False

    # compare will fail if port is provided as string
    if proxyPort is not None and isinstance(proxyPort, basestring):
        proxyPort = int(proxyPort)

    # Create input JSON
    json_data = {
        "config": {
            'enableAutoUpdate': enableAutoUpdate,
            'enableIprAutoUpdate': enableIprAutoUpdate,
            'enableIprFeedback': enableIprFeedback,
            'enableWeblearn': enableWeblearn,
            'includeIprInfo': includeIprInfo,
            'proxyHost': proxyHost,
            'proxyPort': proxyPort,
            'proxyPwd': proxyPwd,
            'proxyUser': proxyUser,
            'useProxy': useProxy,
            'useProxyAuth': useProxyAuth
        }
    }

    # Extract current settings
    ret_obj = get(isamAppliance)

    # Compare to establish idempotency
    sorted_json_data = ibmsecurity.utilities.tools.json_sort(json_data)
    logger.debug("Sorted input: {0}".format(sorted_json_data))
    sorted_ret_obj = ibmsecurity.utilities.tools.json_sort(ret_obj['data'])
    logger.debug("Sorted existing data: {0}".format(sorted_ret_obj))
    if sorted_ret_obj != sorted_json_data:
        logger.info("Changes detected, update needed.")
        update_required = True

    return update_required, json_data


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Application Database Settings between two appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
import ibmsecurity.utilities.tools
import os.path

logger = logging.getLogger(__name__)


def get(isamAppliance, check_mode=False, force=False):
    """
    Get information on existing snapshots
    """
    return isamAppliance.invoke_get("Retrieving snapshots", "/snapshots")


def get_latest(isamAppliance, check_mode=False, force=False):
    """
    Retrieve id of latest found snapshot
    """
    ret_obj_id = isamAppliance.create_return_object()
    ret_obj = get(isamAppliance)

    # Get snapshot with lowest 'id' value - that will be latest one
    snaps = min(ret_obj['data'], key=lambda snap: snap['index'])
    ret_obj_id['data'] = snaps['id']

    return ret_obj_id


def search(isamAppliance, comment, check_mode=False, force=False):
    """
    Retrieve snapshots with given comment contained
    """
    ret_obj = isamAppliance.create_return_object()
    ret_obj_all = get(isamAppliance)

    for obj in ret_obj_all['data']:
        if comment in obj['comment']:
            logger.debug("Snapshot comment \"{0}\" has this string \"{1}\" in it.".format(obj['comment'], comment))
            if ret_obj['data'] == {}:
                ret_obj['data'] = [obj['id']]
            else:
                ret_obj['data'].append(obj['id'])

    return ret_obj


def create(isamAppliance, comment='', check_mode=False, force=False):
    """
    Create a new snapshot
    """
    if force is True or _check(isamAppliance, comment=comment) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post("Creating snapshot", "/snapshots",
                                             {
                                                 'comment': comment
                                             })

    return isamAppliance.create_return_object()


def _check(isamAppliance, comment='', id=None):
    """
    Check if the last created snapshot has the exact same comment or id exists

    :param isamAppliance:
    :param comment:
    :return:
    """
    ret_obj = get(isamAppliance)

    if id != None:
        for snaps in ret_obj['data']:
            if snaps['id'] == id:
                logger.debug("Found id: {}".format(id))
                return True
    else:
        for snaps in ret_obj['data']:
            if snaps['comment'] == comment:
                logger.debug("Found comment: {}".format(comment))
                return True

    return False


def delete(isamAppliance, id=None, comment=None, check_mode=False, force=False):
    """
    Delete snapshot(s) - check id before processing comment. id can be a list
    """
    ids = []
    delete_flag = False
    if (isinstance(id, list)):
        for i in id:
            if _check(isamAppliance, id=i) is True:
                delete_flag = True
                ids.append(i)
    elif (_check(isamAppliance, id=id) is True):
        delete_flag = True
        ids.append(id)
    elif (comment is not None):
        ret_obj = search(isamAppliance, comment=comment)
        if ret_obj != {} and ret_obj['data'] != {}:
            delete_flag = True
            ids = ret_obj['data']
    logger.info("Deleting the following list of IDs: {}".format(ids))
    if force is True or delete_flag is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete("Deleting snapshot",
                                               "/snapshots/multi_destroy?record_ids=" + ",".join(ids))

    return isamAppliance.create_return_object()

def multi_delete(isamAppliance, ids=[], comment=None, check_mode=False, force=False):
    """
    Delete multiple snapshots based on id or comment
    """
    if comment != None:
      ret_obj = search(isamAppliance, comment=comment)
      if ret_obj['data'] == {}:
        return isamAppliance.create_return_object(changed=False)
      else:
        if ids == []:
          ids = ret_obj['data']
        else:
          for snaps in ret_obj['data']:
            ids.append(snaps)

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_delete("Deleting one or multiple snapshots", "/snapshots/multi_destroy?record_ids=" + ",".join(ids))

    return isamAppliance.create_return_object()

def modify(isamAppliance, id, comment, check_mode=False, force=False):
    """
    Modify the snapshot comment
    """
    if force is True or _check(isamAppliance, id=id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put("Modifying snapshot", "/snapshots/" + id,
                                            {
                                                'comment': comment
                                            })

    return isamAppliance.create_return_object()


def apply(isamAppliance, id=None, comment=None, check_mode=False, force=False):
    """
    Apply a snapshot
    There is a priority in the parameter to be used for snapshot applying: id > comment
    """
    apply_flag = False
    if id is not None:
        apply_flag = _check(isamAppliance, id=id)
    elif comment is not None:
        ret_obj = search(isamAppliance, comment)
        if ret_obj['data'] != {}:
            if len(ret_obj['data']) == 1:
                id = ret_obj['data'][0]
                apply_flag = True
            else:
                logger.warn(
                    "There are multiple files with matching comments. Only one snapshot at a time can be applied !")
    else:
        logger.warn("No snapshot detail provided - no id nor comment.")
    if force is True or apply_flag is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post_snapshot_id("Applying snapshot", "/snapshots/apply/" + id,
                                                         {"snapshot_id": id})

    return isamAppliance.create_return_object()


def download(isamAppliance, filename, id=None, comment=None, check_mode=False, force=False):
    """
    Download one snapshot file to a zip file.
    Multiple file download is now supported. Simply pass a list of id.
    For backwards compatibility the id parameter and old behaviour is checked at the beginning.
    """
    ids = []
    download_flag = False
    if (isinstance(id, list)):
        for i in id:
            if _check(isamAppliance, id=i) is True:
                download_flag = True
                ids.append(i)
    elif (_check(isamAppliance, id=id) is True):
        download_flag = True
        ids.append(id)
    elif (comment is not None):
        ret_obj = search(isamAppliance, comment=comment)
        if ret_obj != {} and ret_obj['data'] != {}:
            download_flag = True
            ids = ret_obj['data']
    logger.info("Downloading the following list of IDs: {}".format(ids))

    if force is True or (
            os.path.exists(filename) is False and download_flag is True):  # Don't overwrite if not forced to
        if check_mode is False:  # We are in check_mode but would try to download named ids
            # Download all ids known so far
            return isamAppliance.invoke_get_file("Downloading multiple snapshots",
                                                 "/snapshots/download?record_ids=" + ",".join(ids), filename)

    return isamAppliance.create_return_object()


def download_latest(isamAppliance, dir='.', check_mode=False, force=False):
    """
    Download latest snapshot file to a zip file.
    """
    ret_obj = get(isamAppliance)

    # Get snapshot with lowest 'id' value - that will be latest one
    snaps = min(ret_obj['data'], key=lambda snap: snap['index'])
    id = snaps['id']
    file = snaps['filename']
    filename = os.path.join(dir, file)

    return download(isamAppliance, filename, id, check_mode, force)


def apply_latest(isamAppliance, check_mode=False, force=False):
    """
    Apply latest snapshot file (revert to latest)
    """
    ret_obj = get(isamAppliance)

    # Get snapshot with lowest 'id' value - that will be latest one
    snaps = min(ret_obj['data'], key=lambda snap: snap['index'])
    id = snaps['id']

    return apply(isamAppliance, id, check_mode, force)


def upload(isamAppliance, file, comment=None, check_mode=False, force=False):
    """
    Upload Snapshot file
    """
    if comment is None:
        import zipfile

        zFile = zipfile.ZipFile(file)
        if "Comment" in zFile.namelist():
            comment = zFile.open("Comment")

    if force is True or _check(isamAppliance, comment=comment) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:

            return isamAppliance.invoke_post_files(
                "Upload Snapshot",
                "/snapshots",
                [{
                    'file_formfield': 'uploadedfile',
                    'filename': file,
                    'mimetype': 'application/octet-stream'
                }],
                {
                    'comment': comment if comment != None else ''
                }, json_response=False)

    return isamAppliance.create_return_object()


def compare(isamAppliance1, isamAppliance2):
    """
    Compare list of snapshots between 2 appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    # id of snapshot is uniquely generated on appliance and should therefore be ignored in comparison.
    # filename of snapshot is generated based on exact date/time and will differ even if 2 snapshots are taken near the
    # same time. Therefore, filename should be ignored in comparison
    for snapshot in ret_obj1['data']:
        del snapshot['id']
        del snapshot['filename']

    for snapshot in ret_obj2['data']:
        del snapshot['id']
        del snapshot['filename']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id', 'filename'])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/wga/redis_config/collections"
requires_modules = ["wga"]
requires_version = "10.0.1"


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve the list of configured Redis collections
    """
    return isamAppliance.invoke_get("Retrieve the list of configured Redis collections",
                                    "{0}".format(uri),
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve the list of configured Redis collections
    """
    return isamAppliance.invoke_get("Retrieve the list of configured Redis collections",
                                    "{0}/{1}".format(uri, name),
                                    requires_modules=requires_modules, requires_version=requires_version)


def add(isamAppliance, name, max_pooled_connections=50, idle_timeout=10, connect_timeout=5,
        io_timeout=30, health_check_interval=15, servers=[], cross_domain_support=None, matching_hosts=None,
        check_mode=False, force=False):
    """
    Create the configuration for a new collection of Redis servers
    """

    exist, warnings = _check_exist(isamAppliance, name)

    if force is True or exist is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = {
                "name": name,
                "max-pooled-connections": max_pooled_connections,
                "idle-timeout": idle_timeout,
                "connect-timeout": connect_timeout,
                "io-timeout": io_timeout,
                "health-check-interval": health_check_interval,
                "servers": servers
            }

            if matching_hosts != None:
                json_data['matching-hosts'] = matching_hosts

            if cross_domain_support != None:
                json_data['cross-domain-support'] = cross_domain_support

            return isamAppliance.invoke_post(
                "Create the configuration for a new collection of Redis servers",
                "{0}".format(uri),
                json_data,
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def update(isamAppliance, name, max_pooled_connections=50, idle_timeout=10, connect_timeout=5,
           io_timeout=30, health_check_interval=15, servers=None, cross_domain_support=None, matching_hosts=None,
           check_mode=False, force=False):
    """
    Update the configuration of a collection of Redis servers
    """

    exist, warnings = _check_exist(isamAppliance, name)
    if exist is True:
        same_contents, warnings, json_data = _check_contents(isamAppliance=isamAppliance, name=name,
                                                             max_pooled_connections=max_pooled_connections,
                                                             idle_timeout=idle_timeout, connect_timeout=connect_timeout,
                                                             io_timeout=io_timeout,
                                                             health_check_interval=health_check_interval,
                                                             servers=servers, cross_domain_support=cross_domain_support,
                                                             matching_hosts=matching_hosts)
    else:
        return isamAppliance.create_return_object(warnings=warnings)

    if force is True or same_contents is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            if 'servers' not in json_data:
                json_data['servers'] = []
            return isamAppliance.invoke_put(
                "Update the configuration of a collection of Redis servers",
                "{0}/{1}".format(uri, name),
                json_data,
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def set(isamAppliance, name, max_pooled_connections=50, idle_timeout=10, connect_timeout=5,
        io_timeout=30, health_check_interval=15, servers=None, cross_domain_support=None, matching_hosts=None,
        check_mode=False, force=False):
    exist, warnings = _check_exist(isamAppliance, name)
    if exist is True:
        return update(isamAppliance=isamAppliance, name=name, max_pooled_connections=max_pooled_connections,
                      idle_timeout=idle_timeout, connect_timeout=connect_timeout, io_timeout=io_timeout,
                      health_check_interval=health_check_interval, servers=servers,
                      cross_domain_support=cross_domain_support, matching_hosts=matching_hosts, check_mode=check_mode,
                      force=force)
    else:
        if servers is None:
            servers = []
        return add(isamAppliance=isamAppliance, name=name, max_pooled_connections=max_pooled_connections,
                   idle_timeout=idle_timeout, connect_timeout=connect_timeout, io_timeout=io_timeout,
                   health_check_interval=health_check_interval, servers=servers,
                   cross_domain_support=cross_domain_support, matching_hosts=matching_hosts, check_mode=check_mode,
                   force=force)


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete the configuration for a collection of Redis servers
    """
    exists, warnings = _check_exist(isamAppliance, name)

    if force is True or exists is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete the configuration for a collection of Redis servers",
                "{0}/{1}".format(uri, name),
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def _check_exist(isamAppliance, name):
    ret_obj = get_all(isamAppliance)

    for obj in ret_obj['data']:
        if obj['name'] == name:
            return True, ret_obj['warnings']

    return False, ret_obj['warnings']


def _check_contents(isamAppliance, name, max_pooled_connections, idle_timeout, connect_timeout, io_timeout,
                    health_check_interval, servers, cross_domain_support, matching_hosts):
    ret_obj = get(isamAppliance, name)
    current_content = ret_obj['data']
    warnings = ret_obj['warnings']

    json_data = {
        'name': name,
        'max-pooled-connections': max_pooled_connections,
        'idle-timeout': idle_timeout,
        'connect-timeout': connect_timeout,
        'io-timeout': io_timeout,
        'health-check-interval': health_check_interval
    }
    if servers is not None:
        json_data['servers'] = servers

    if cross_domain_support is not None:
        json_data['cross-domain-support'] = cross_domain_support

    if matching_hosts is not None:
        json_data['matching-hosts'] = matching_hosts

    sorted_obj1 = tools.json_sort(json_data)
    logger.debug("Sorted sorted_obj1: {0}".format(sorted_obj1))
    sorted_obj2 = tools.json_sort(current_content)
    logger.debug("Sorted sorted_obj2: {0}".format(sorted_obj2))

    if sorted_obj1 != sorted_obj2:
        logger.info("Changes detected, update needed.")
        return False, warnings, json_data
    else:
        return True, warnings, json_data


def compare(isamAppliance1, isamAppliance2):
    """
    Compare redis configurations between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    return tools.json_compare(ret_obj1, ret_obj2)



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)

uri = "/isam/pdadmin"
requires_version = "10.0.0"
requires_modules = None
requires_model = "Appliance"
version = "v1"

def retrieve(isamAppliance, admin_id, admin_pwd, object='/', admin_domain='Default', check_mode=False, force=False):
    """
    Retrieve a list of objects
    Fix: do not consider 404 an error
    """
    _action = "objectlistandshowext"
    ret_obj = isamAppliance.invoke_post(description="Retrieve a list of objects",
                                        uri=f"{uri}/{_action}/{version}",
                                        ignore_error=True,
                                        data={
                                            "admin_id": admin_id,
                                            "admin_pwd": admin_pwd,
                                            "admin_domain": admin_domain,
                                            "object": object
                                        })
    ret_obj['changed'] = False
    if ret_obj['rc'] == 404:
        logger.info(f"Object {object} could not be found in {admin_domain} for list and show")
        return isamAppliance.create_return_object()
    return ret_obj

def get(isamAppliance, admin_id, admin_pwd, object='/', admin_domain='Default', check_mode=False, force=False):
    """
    Retrieve a specific object
    Fix: do not consider 404 an error
    """
    _action = "objectshowext"
    ret_obj = isamAppliance.invoke_post(description="Retrieve a specific object",
                                        uri=f"{uri}/{_action}/{version}",
                                        ignore_error=True,
                                        data={
                                            "admin_id": admin_id,
                                            "admin_pwd": admin_pwd,
                                            "admin_domain": admin_domain,
                                            "object": object
                                        })
    ret_obj['changed'] = False
    if ret_obj['rc'] == 404:
        logger.info(f"Object {object} could not be found in {admin_domain}")
        return isamAppliance.create_return_object()

    return ret_obj


def compare(isamAppliance1, isamAppliance2, admin_id, admin_pwd, admin_domain='Default', check_mode=False, force=False):
    """
    Compare objects between two appliances
    Note that this only compares the first level
    """
    ret_obj1 = retrieve(isamAppliance1, admin_id, admin_pwd, '/', admin_domain)
    ret_obj2 = retrieve(isamAppliance2, admin_id, admin_pwd, '/', admin_domain)

    for obj in ret_obj1['data']:
        ret_obj = get(isamAppliance1, admin_id, admin_pwd, object=obj['id'], admin_domain=admin_domain)
        obj['script'] = ret_obj['data']['contents']
    for obj in ret_obj2['data']:
        ret_obj = get(isamAppliance2, admin_id, admin_pwd, object=obj['id'], admin_domain=admin_domain)
        obj['script'] = ret_obj['data']['contents']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
from ibmsecurity.utilities import tools

try:
    basestring
except NameError:
    basestring = (str, bytes)

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/audit"
requires_modules = ["mga", "federation"]
requires_version = None


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve audit configuration
    """
    return isamAppliance.invoke_get("Retrieve audit configuration", uri, requires_modules=requires_modules,
                                    requires_version=requires_version)


def set(isamAppliance, id, config, enabled=True, type='Syslog', verbose=True, check_mode=False, force=False, use_json=False, components=None):
    """
    Update Audit Configuration

    Sample data for Audit Configuration:
    In JSON Format:
    { u'config': [ { u'datatype': u'String',
                     u'key': u'ISAM.Audit.syslogclient.SSL_TRUST_STORE',
                     u'sensitive': False,
                     u'validValues': [],
                     u'value': u''},
                   { u'datatype': u'String',
                     u'key': u'ISAM.Audit.syslogclient.CLIENT_AUTH_KEY',
                     u'sensitive': False,
                     u'validValues': [],
                     u'value': u''},
                   { u'datatype': u'Boolean',
                     u'key': u'ISAM.Audit.syslogclient.FAILOVER_TO_DISK',
                     u'sensitive': False,
                     u'validValues': [],
                     u'value': u'false'},
                   { u'datatype': u'Integer',
                     u'key': u'ISAM.Audit.syslogclient.NUM_RETRY',
                     u'sensitive': False,
                     u'validValues': [],
                     u'value': u'2'},
                   { u'datatype': u'Integer',
                     u'key': u'ISAM.Audit.syslogclient.NUM_SENDER_THREADS',
                     u'sensitive': False,
                     u'validValues': [],
                     u'value': u'1'},
                   { u'datatype': u'Boolean',
                     u'key': u'ISAM.Audit.syslogclient.CLIENT_CERT_AUTH_REQUIRED',
                     u'sensitive': False,
                     u'validValues': [],
                     u'value': u'false'},
                   { u'datatype': u'Integer',
                     u'key': u'ISAM.Audit.syslogclient.SERVER_PORT',
                     u'sensitive': False,
                     u'validValues': [],
                     u'value': u'514'},
                   { u'datatype': u'Hostname',
                     u'key': u'ISAM.Audit.syslogclient.SERVER_HOST',
                     u'sensitive': False,
                     u'validValues': [],
                     u'value': u'127.0.0.1'},
                   { u'datatype': u'String',
                     u'key': u'ISAM.Audit.syslogclient.TRANSPORT',
                     u'sensitive': False,
                     u'validValues': [],
                     u'value': u'TRANSPORT_UDP'},
                   { u'datatype': u'Integer',
                     u'key': u'ISAM.Audit.syslogclient.QUEUE_FULL_TIMEOUT',
                     u'sensitive': False,
                     u'validValues': [],
                     u'value': u'-1'},
                   { u'datatype': u'Integer',
                     u'key': u'ISAM.Audit.syslogclient.MAX_QUEUE_SIZE',
                     u'sensitive': False,
                     u'validValues': [],
                     u'value': u'1000'}],
      u'enabled': False,
      u'id': u'1',
      u'type': u'Syslog',
      u'verbose': False}

    In YAML Format:
    config:
      - datatype: String
        key: ISAM.Audit.syslogclient.SSL_TRUST_STORE
        sensitive: false
        validValues: []
        value: ''
      - datatype: String
        key: ISAM.Audit.syslogclient.CLIENT_AUTH_KEY
        sensitive: false
        validValues: []
        value: ''
      - datatype: Boolean
        key: ISAM.Audit.syslogclient.FAILOVER_TO_DISK
        sensitive: false
        validValues: []
        value: 'false'
      - datatype: Integer
        key: ISAM.Audit.syslogclient.NUM_RETRY
        sensitive: false
        validValues: []
        value: '2'
      - datatype: Integer
        key: ISAM.Audit.syslogclient.NUM_SENDER_THREADS
        sensitive: false
        validValues: []
        value: '1'
      - datatype: Boolean
        key: ISAM.Audit.syslogclient.CLIENT_CERT_AUTH_REQUIRED
        sensitive: false
        validValues: []
        value: 'false'
      - datatype: Integer
        key: ISAM.Audit.syslogclient.SERVER_PORT
        sensitive: false
        validValues: []
        value: '514'
      - datatype: Hostname
        key: ISAM.Audit.syslogclient.SERVER_HOST
        sensitive: false
        validValues: []
        value: 127.0.0.1
      - datatype: String
        key: ISAM.Audit.syslogclient.TRANSPORT
        sensitive: false
        validValues: []
        value: TRANSPORT_UDP
      - datatype: Integer
        key: ISAM.Audit.syslogclient.QUEUE_FULL_TIMEOUT
        sensitive: false
        validValues: []
        value: '-1'
      - datatype: Integer
        key: ISAM.Audit.syslogclient.MAX_QUEUE_SIZE
        sensitive: false
        validValues: []
        value: '1000'
    enabled: false
    id: '1'
    type: Syslog
    verbose: false
    """
    pol_id, update_required, json_data = _check(isamAppliance, id, config, enabled, type, verbose, use_json, components)
    if pol_id is None:
        from ibmsecurity.appliance.ibmappliance import IBMError
        raise IBMError("999", "Cannot update data for unknown Audit Configuration ID: {0}".format(id))

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update Audit Configuration",
                "{0}/{1}".format(uri, id), json_data, requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def _check(isamAppliance, id, config, enabled, type, verbose, use_json=False, components=None):
    """
    Check and return True if update needed
    """
    update_required = False
    pol_id = None
    # convert all values into string - any other type causes issues
    for cfg in config:
        if isinstance(cfg['value'], bool):
            cfg['value'] = str(cfg['value']).lower()
        else:
            cfg['value'] = str(cfg['value'])
    # Ensure boolean variables are set correctly
    if isinstance(verbose, str):
        if verbose.lower() == "true":
            verbose = True
        else:
            verbose = False
    if isinstance(enabled, str):
        if enabled.lower() == "true":
            enabled = True
        else:
            enabled = False
    if isinstance(use_json, str):
        if use_json.lower() == "true":
            use_json = True
        else:
            use_json = False
    json_data = {
        "id": id,
        "config": config,
        "enabled": enabled,
        "type": type,
        "verbose": verbose,
        "useJSONFormat": use_json
    }
    ret_obj = get(isamAppliance)
    for aud_cfg in ret_obj['data']:
        if id == aud_cfg['id']:
            pol_id = id
            break
    if pol_id is None:
        logger.warning("Audit Configuration not found, returning no update required.")
        return pol_id, update_required, json_data
    elif components is not None:
        json_data["components"] = components
        update_required = True
    else:
        import ibmsecurity.utilities.tools
        sorted_json_data = ibmsecurity.utilities.tools.json_sort(json_data)
        logger.debug("Sorted input: {0}".format(sorted_json_data))
        sorted_ret_obj = ibmsecurity.utilities.tools.json_sort(aud_cfg)
        logger.debug("Sorted existing data: {0}".format(sorted_ret_obj))
        if sorted_ret_obj != sorted_json_data:
            logger.info("Changes detected, update needed.")
            update_required = True

    return pol_id, update_required, json_data


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Audit Configuration between two appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['id']
    for obj in ret_obj2['data']:
        del obj['id']

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id'])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/devicefingerprints/"
requires_modules = ["mga"]
requires_version = None


def get_all(isamAppliance, filter=None, sortBy=None, check_mode=False, force=False):
    """
    Retrieve a list of devices fingerprints
    """
    return isamAppliance.invoke_get("Retrieve a list of Device Fingerprints",
                                    "{0}/{1}".format(uri, tools.create_query_string(filter=filter, sortBy=sortBy)),
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a device fingerprint for a specific device
    """
    warnings = []
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    id = ret_obj['data']

    if id == {}:
        warnings.append("Device {0} had no match, skipping retrieval.".format(name))
        return isamAppliance.create_return_object(changed=False, warnings=warnings)
    else:
        return isamAppliance.invoke_get("Retrieve a specific Device Fingerprint",
                                        "{0}/{1}".format(uri, id))


def search(isamAppliance, name, force=False, check_mode=False):
    """
    Search device id by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['name'] == name:
            logger.info("Found Device {0} id: {1}".format(name, obj['id']))
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete a device fingerprint for a specific device
    """
    warnings = []
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    id = ret_obj['data']

    if id == {}:
        warnings.append("Device {0} had no match, skipping delete.".format(name))
        return isamAppliance.create_return_object(changed=False, warnings=warnings)
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete("Delete a specific Device Fingerprint",
                                               "{0}/{1}".format(uri, id))


def delete_set(isamAppliance, devices, check_mode=False, force=False):
    """
    Delete a set of device fingerprints
    """
    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_delete(description="Delete a set of device fingerprints",
                                           uri=uri, data=devices, ignore_error=True)



import json
import time
import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get(isdsAppliance, check_mode=False, force=False):
    """
    Get current configured server type
    """
    return isdsAppliance.invoke_get("Retrieving Server Status", "/widgets/server")


def start(isdsAppliance, serverID='directoryserver', check_mode=False, force=False):
    """
    Restart the specified appliance server
    """
    if force is True or _check(isdsAppliance, serverID, action='start') is True:
        if check_mode is True:
            return isdsAppliance.create_return_object(changed=True)
        else:
            return isdsAppliance.invoke_post("Restarting the service " + serverID,
                                             "/widgets/server/start/" + serverID,
                                             {})

    return isdsAppliance.create_return_object()


def startconfig(isdsAppliance, serverID='directoryserver', check_mode=False, force=False):
    """
    Restart the specified appliance server
    """
    if force is True or _check(isdsAppliance, serverID, action='startconfig') is True:
        if check_mode is True:
            return isdsAppliance.create_return_object(changed=True)
        else:
            return isdsAppliance.invoke_post("Restarting the service " + serverID,
                                             "/widgets/server/startconfig/" + serverID,
                                             {})

    return isdsAppliance.create_return_object()


def stop(isdsAppliance, serverID='directoryserver', check_mode=False, force=False):
    """
    Restart the specified appliance server
    """
    if force is True or _check(isdsAppliance, serverID, action='stop') is True:
        if check_mode is True:
            return isdsAppliance.create_return_object(changed=True)
        else:
            return isdsAppliance.invoke_post("Restarting the service " + serverID,
                                             "/widgets/server/stop/" + serverID,
                                             {})

    return isdsAppliance.create_return_object()


def restart(isdsAppliance, serverID='directoryserver', check_mode=False, force=False):
    """
    Restart the specified appliance server
    """
    if force is True or _check(isdsAppliance, serverID, action='restart') is True:
        if check_mode is True:
            return isdsAppliance.create_return_object(changed=True)
        else:
            return isdsAppliance.invoke_post("Restarting the service " + serverID,
                                             "/widgets/server/restart/" + serverID,
                                             {})

    return isdsAppliance.create_return_object()


def _check(isdsAppliance, serverID, action):
    """
    Check if serverID one of these acceptable values:
        directoryserver
        directoryadminserver
        directorywat
        directoryintegrator
        directoryintegratorscimtarget
        scimservice
    Note: only directoryserver supports "startconfig" action
    """

    if serverID == 'directoryserver':
        if action == 'startconfig':
            return True
        return True
    elif serverID == 'directoryadminserver':
        if action == 'startconfig':
            return False
        return True
    elif serverID == 'directorywat':
        if action == 'startconfig':
            return False
        return True
    elif serverID == 'directoryintegrator':
        if action == 'startconfig':
            return False
        return True
    elif serverID == 'directoryintegratorscimtarget':
        if action == 'startconfig':
            return False
        return True
    elif serverID == 'scimservice':
        if action == 'startconfig':
            return False
        return True
    else:
        return False



import logging
from io import open

logger = logging.getLogger(__name__)

# URI for this module
uri = "/core/licenses"
requires_modules = None
requires_version = None


def install(isamAppliance, license, check_mode=False, force=False):
    """
    install license in the appliance
    license should be the filename of the license file
    """
    # create the request header for the post first
    headers = {
        "Accept": "text/html"
    }

    file = {"license": open(license, 'rb')}

    if _check_license(isamAppliance, license) == True and force == False:
        return isamAppliance.create_return_object(warnings=["License already installed"])
    else:
        if _check_license(isamAppliance, license) == False or force == True:
            if check_mode:
                return isamAppliance.create_return_object(
                    changed=True)  # there is no real change in appliance, so maybe it should be false
            else:
                return isamAppliance.invoke_request("Applying license to appliance", method="post", uri=uri,
                                                    requires_modules=requires_modules,
                                                    requires_version=requires_version,
                                                    warnings=[],
                                                    headers=headers, files=file)

    return isamAppliance.create_return_object()


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve all licenses installed in the appliance
    """
    return isamAppliance.invoke_get("Retrieve all licenses installed in the appliance",
                                    "{0}".format(uri), requires_modules=requires_modules,
                                    requires_version=requires_version)


def _check_license(isamAppliance, license):
    """
    check if a particular license file is installed in the appliance
    """

    ret_obj = get_all(isamAppliance)

    if ret_obj['data'] == {}:
        return False  # no license installed in the appliance

    from xml.etree import ElementTree

    with open(license, 'rt') as f:
        tree = ElementTree.parse(f)

    for path in ['.//OCN']:
        node = tree.find(path)
        if node is not None:
            ocnnumber = node.text
            for item, value in ret_obj['data'].items():
                if 'ocn' in value and value['ocn'] == ocnnumber:
                    return True
        else:
            return False  # does not look like a valid input license file

    return False


def compare(isamAppliance1, isamAppliance2):
    """
    Compare license installed between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['serial_number'])



# some shared variables.
# Quick & dirty
server_fields = {'server_hostname': {'type': 'string'},
                 'server_port': {'type': 'number'},
                 'case_sensitive_url': {'type': 'yesno', 'max_version': "10.0.6.0"},
                 'case_insensitive_url': {'type': 'yesno', 'min_version': "10.0.6.0"},
                 'http_port': {'type': 'number'},
                 'local_ip': {'type': 'string'},
                 'query_contents': {'type': 'string', 'alt_name': 'query_content_url'},
                 'server_dn': {'type': 'string'},
                 'server_uuid': {'type': 'ignore'},
                 'virtual_hostname': {'type': 'string', 'alt_name': 'virtual_junction_hostname'},
                 'windows_style_url': {'type': 'yesno'},
                 'current_requests': {'type': 'ignore'},
                 'total_requests': {'type': 'ignore'},
                 'operation_state': {'type': 'ignore'},
                 'server_state': {'type': 'ignore'},
                 'priority': {'type': 'number', 'min_version': "10.0.2.0"},
                 'server_cn': {'type': 'string', 'min_version': "10.0.2.0"}}


import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/core/update_servers"
requires_modules = None
requires_version = None


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Get all Update Servers
    """
    return isamAppliance.invoke_get("Get Update Servers", uri, requires_modules=requires_modules,
                                    requires_version=requires_version)


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a specific update server
    """
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    us_id = ret_obj['data']

    if us_id == {}:
        logger.info("Update Server {0} had no match, skipping retrieval.".format(name))
        return isamAppliance.create_return_object()
    else:
        return _get(isamAppliance, us_id)


def _get(isamAppliance, us_id):
    return isamAppliance.invoke_get("Retrieve a specific update server",
                                    "{0}/{1}".format(uri, us_id))


def search(isamAppliance, name, force=False, check_mode=False):
    """
    Search update server by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj['data']['luServers']:
        if obj['name'] == name:
            logger.info("Found Update Server {0} id: {1}".format(name, obj['uuid']))
            return_obj['data'] = obj['uuid']
            return_obj['rc'] = 0

    return return_obj


def set(isamAppliance, priority, name, enabled, hostName, port, trustLevel, useProxy=False, useProxyAuth=False,
        cert=None, proxyHost=None, proxyPort=None, proxyUser=None, proxyPwd=None, new_name=None, check_mode=False,
        force=False):
    """
    Creating or Modifying a update server
    """
    if (search(isamAppliance, name=name))['data'] == {}:
        # Force the add - we already know update server does not exist
        logger.info("Update Server {0} had no match, requesting to add new one.".format(name))
        return add(isamAppliance, priority, name, enabled, hostName, port, trustLevel, useProxy,
                   useProxyAuth, cert, proxyHost, proxyPort, proxyUser, proxyPwd, check_mode, True)
    else:
        # Update request
        logger.info("Update Server {0} exists, requesting to update.".format(name))
        return update(isamAppliance, priority, name, enabled, hostName, port, trustLevel, useProxy, useProxyAuth, cert,
                      proxyHost, proxyPort, proxyUser, proxyPwd, new_name, check_mode, force)


def add(isamAppliance, priority, name, enabled, hostName, port, trustLevel, useProxy=False, useProxyAuth=False, cert="",
        proxyHost=None, proxyPort=None, proxyUser=None, proxyPwd=None, check_mode=False, force=False):
    """
    Add a Update Server
    """
    if force is True or search(isamAppliance, name=name) == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = {"priority": priority,
                         "name": name,
                         "enabled": enabled,
                         "hostName": hostName,
                         "port": port,
                         "trustLevel": trustLevel,
                         "useProxy": useProxy,
                         "useProxyAuth": useProxyAuth,
                         "_isNew": True,
                         "cert": cert,
                         "proxyHost": proxyHost,
                         "proxyPort": proxyPort,
                         "proxyUser": proxyUser,
                         "proxyPwd": proxyPwd}

            return isamAppliance.invoke_post("Add a Update Server", uri, json_data, requires_modules=requires_modules,
                                             requires_version=requires_version)

    return isamAppliance.create_return_object()


def update(isamAppliance, priority, name, enabled, hostName, port, trustLevel, useProxy=False, useProxyAuth=False,
           cert="", proxyHost=None, proxyPort=None, proxyUser=None, proxyPwd=None, new_name=None, check_mode=False,
           force=False):
    """
    Update an update server's details.
    """
    us_id, update_required, json_data = _check(isamAppliance, priority, name, enabled, hostName, port, trustLevel,
                                               useProxy, useProxyAuth, cert, proxyHost, proxyPort, proxyUser, proxyPwd,
                                               new_name)

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put("Update an update server", "{0}/{1}".format(uri, us_id), json_data,
                                            requires_modules=requires_modules,
                                            requires_version=requires_version)

    return isamAppliance.create_return_object()


def _check(isamAppliance, priority, name, enabled, hostName, port, trustLevel, useProxy, useProxyAuth,
           cert, proxyHost, proxyPort, proxyUser, proxyPwd, new_name):
    update_required = False
    json_data = {"priority": priority,
                 "name": name,
                 "enabled": enabled,
                 "hostName": hostName,
                 "port": port,
                 "trustLevel": trustLevel,
                 "useProxy": useProxy,
                 "useProxyAuth": useProxyAuth,
                 "cert": cert,
                 "proxyHost": proxyHost,
                 "proxyPort": proxyPort,
                 "proxyUser": proxyUser,
                 "proxyPwd": proxyPwd}
    ret_obj = get(isamAppliance, name)
    if ret_obj['data'] == {}:
        logger.warning("Update Server not found, returning no update required.")
        return None, update_required, json_data
    else:
        us_id = ret_obj['data']['uuid']
        if new_name is not None:
            json_data['name'] = new_name
        else:
            json_data['name'] = name
        del ret_obj['data']['uuid']
        sorted_json_data = ibmsecurity.utilities.tools.json_sort(json_data)
        logger.debug("Sorted input: {0}".format(sorted_json_data))
        sorted_ret_obj = ibmsecurity.utilities.tools.json_sort(ret_obj['data'])
        logger.debug("Sorted existing data: {0}".format(sorted_ret_obj))
        if sorted_ret_obj != sorted_json_data:
            logger.info("Changes detected, update needed.")
            update_required = True

    return us_id, update_required, json_data


def enable(isamAppliance, name, enabled, check_mode=False, force=False):
    """
    Update an update server's details.
    """
    warnings = []
    ret_obj = get(isamAppliance=isamAppliance, name=name)

    if ret_obj['data'] == {}:
        warnings.append("Update Server {0} not found.".format(name))
    elif force is True or enabled != ret_obj['data']['enabled']:
        logger.debug("Enable flag needs to be updated!")
        return update(isamAppliance, priority=ret_obj['data']['priority'], name=name, enabled=enabled,
                      hostName=ret_obj['data']['hostName'], port=ret_obj['data']['port'],
                      trustLevel=ret_obj['data']['trustLevel'], useProxy=ret_obj['data']['useProxy'],
                      useProxyAuth=ret_obj['data']['useProxyAuth'], cert=ret_obj['data']['cert'],
                      proxyHost=ret_obj['data']['proxyHost'], proxyPort=ret_obj['data']['proxyPort'],
                      proxyUser=ret_obj['data']['proxyUser'], proxyPwd=ret_obj['data']['proxyPwd'],
                      check_mode=check_mode,
                      force=True)

    return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete an Update Server
    """
    ret_obj = search(isamAppliance, name=name)

    if ret_obj['data'] != {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete("Delete an Update Server", "{0}/{1}".format(uri, ret_obj['data']))
    else:
        logger.info("Update Server: {0} not found, delete skipped.".format(name))

    return isamAppliance.create_return_object()


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Update Servers between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['uuid']
    for obj in ret_obj2['data']:
        del obj['uuid']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['uuid'])



import logging
import json
from ibmsecurity.utilities import tools
from io import open

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/federations"
requires_modules = ["federation"]
requires_version = "9.0.1.0"


def get_all(isamAppliance, count=None, start=None, filter=None, check_mode=False, force=False):
    """
    Retrieve a list of federations
    """
    return isamAppliance.invoke_get("Retrieve a list of federations",
                                    "{0}/{1}".format(uri, tools.create_query_string(count=count, start=start,
                                                                                    filter=filter)),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a federation
    """
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    id = ret_obj['data']

    if id == {}:
        logger.info("Federation {0} had no match, skipping retrieval.".format(name))
        return isamAppliance.create_return_object()
    else:
        return _get(isamAppliance, id)


def get_templates(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve the templates for a federation
    """
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    id = ret_obj['data']

    if id == {}:
        logger.info("Federation {0} had no match, skipping templates retrieval.".format(name))
        return isamAppliance.create_return_object()
    else:
        return isamAppliance.invoke_get("Retrieve the templates for a federation",
                                        "{0}/{1}/templates".format(uri, id),
                                        requires_modules=requires_modules,
                                        requires_version=requires_version)


def _get(isamAppliance, id):
    """
    Internal function to get data using "id" - used to avoid extra calls

    :param isamAppliance:
    :param id:
    :return:
    """
    return isamAppliance.invoke_get("Retrieve a specific federation",
                                    "{0}/{1}".format(uri, id),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def set(isamAppliance, name, protocol, configuration, role=None, templateName=None, new_name=None, check_mode=False,
        force=False):
    """
    Creating or Modifying a Federation
    """
    if (search(isamAppliance, name=name))['data'] == {}:
        # Force the add - we already know federation does not exist
        isamAppliance.logger.info("Federation {0} had no match, requesting to add new one.".format(name))
        return add(isamAppliance, name=name, protocol=protocol, role=role, configuration=configuration,
                   templateName=templateName, check_mode=check_mode, force=True)
    else:
        # Update request
        isamAppliance.logger.info("Federation {0} exists, requesting to update.".format(name))
        return update(isamAppliance, name=name, protocol=protocol, role=role, configuration=configuration, templateName=templateName,
                      new_name=new_name, check_mode=check_mode, force=force)


def set_file(isamAppliance, name, protocol, filename, mapping_id, role=None, templateName=None, new_name=None,
             check_mode=False, force=False):
    """
    Creating or Modifying a Federation from a JSON file
    """
    with open(filename, 'r') as infile:
        file_lines = json.load(infile)

    # update mapping_id
    if file_lines.get('configuration', None) is not None:
        file_lines = file_lines['configuration']
    file_lines['identityMapping']['properties']['identityMappingRuleReference'] = str(mapping_id)

    return set(isamAppliance=isamAppliance, name=name, protocol=protocol, configuration=file_lines, role=role,
               templateName=templateName, new_name=new_name, check_mode=check_mode, force=force)


def add(isamAppliance, name, protocol, configuration, role=None, templateName=None, check_mode=False,
        force=False):
    """
    Create a new federation

    :param isamAppliance:
    :param name:
    :param protocol: SAML2_0 or OIDC or OIDC10 (OIDC deprecated on v10) or WSFED
    :param role: ip | sp | op | rp (not required on OIDC10)
    :param configuration: protocol specific configuration data in JSON format
    :param templateName: template id (optional)
    :param check_mode:
    :param force:
    :return:
    """
    if force is False:
        ret_obj = search(isamAppliance, name)

    if force is True or ret_obj['data'] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = {
                "name": name,
                "protocol": protocol,
                "configuration": configuration
            }
            if role is not None:
                json_data['role'] = role
            if templateName is not None:
                json_data['templateName'] = templateName
            return isamAppliance.invoke_post(
                "Create a new federation",
                uri, json_data,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete a federation
    """
    ret_obj = search(isamAppliance, name, check_mode=check_mode, force=force)
    fed_id = ret_obj['data']

    if fed_id == {}:
        logger.info("Federation {0} not found, skipping delete.".format(name))
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a federation",
                "{0}/{1}".format(uri, fed_id),
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def update(isamAppliance, name, role=None, protocol=None, configuration=None, templateName=None, new_name=None, check_mode=False,
           force=False):
    """
    Update a specific federation
    """
    fed_id, update_required, json_data = _check(isamAppliance, name, role, protocol, configuration, templateName, new_name)
    if fed_id is None:
        from ibmsecurity.appliance.ibmappliance import IBMError
        raise IBMError("999", "Cannot update data for unknown federation: {0}".format(name))

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specific federation",
                "{0}/{1}".format(uri, fed_id), json_data,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def _check(isamAppliance, name, role, protocol, configuration, templateName=None, new_name=None):
    """
    Check and return True if update needed

    :param isamAppliance:
    :param name:
    :param role:
    :param protocol:
    :param configuration:
    :param templateName:
    :return:
    """
    update_required = False
    json_data = {}
    ret_obj = get(isamAppliance, name)
    if ret_obj['data'] == {}:
        logger.info("Federation not found, returning no update required.")
        return None, update_required, json_data
    else:
        fed_id = ret_obj['data']['id']
        if new_name is not None:
            json_data['name'] = new_name
        else:
            json_data['name'] = name
        # Only added parameters passed, remove everything else
        if role is not None:
            json_data['role'] = role
        else:
            del ret_obj['data']['role']
        if templateName is not None:
            json_data['templateName'] = templateName
        else:
            ret_obj['data'].pop('templateName')
        if configuration is not None:
            json_data['configuration'] = configuration
            # Check to see if configuration data contains mapping rule reference id
            # So special logic to see if mapping rule has changed
            # TODO: WHY ????
            new_map_rule_id, new_map_rule = None, None
            exist_map_rule_id, exist_map_rule = None, None

            new_map_rule_id = configuration['identityMapping']['properties'].get('identityMappingRuleReference', None)
            new_map_rule = configuration['identityMapping']['properties'].get('identityMappingRule', None)

            exist_map_rule_id = ret_obj['data']['configuration']['identityMapping']['properties'].get('identityMappingRuleReference', None)
            exist_map_rule = ret_obj['data']['configuration']['identityMapping']['properties'].get('identityMappingRule', None)

            logger.debug("New Mapping data: {0}/{1}".format(new_map_rule_id, new_map_rule))
            logger.debug("Existing Mapping data: {0}/{1}".format(exist_map_rule_id, exist_map_rule))
            if new_map_rule_id is not None:
                # Mapping rule id provided in update, check if we need to extract actual rules and compare
                if exist_map_rule_id is None and exist_map_rule is not None:
                    import ibmsecurity.isam.aac.mapping_rules
                    new_map_obj = ibmsecurity.isam.aac.mapping_rules._get(isamAppliance, new_map_rule_id)
                    if new_map_obj['data'] != {}:
                        if new_map_obj['data']['content'].strip() != exist_map_rule.strip():
                            logger.info("Mapping Rule is different, need to update.")
                            return fed_id, True, json_data
                # Allow mapping rule id to be compared - not actual rules contents
                elif exist_map_rule_id is not None:
                    logger.debug("Comparing mapping rule ids and ignoring actual rule contents.")
                    if exist_map_rule is not None:
                        del ret_obj['data']['configuration']['identityMapping']['properties']['identityMappingRule']
                    if new_map_rule is not None:
                        del json_data['configuration']['identityMapping']['properties']['identityMappingRule']
        else:
            json_data['configuration'] = {}
        # Potential for missing mapping rule - so add configuration back
        if configuration is not None:
            json_data['configuration'] = configuration

        ret_obj['data'].pop('id')
        if protocol is not None:
            json_data['protocol'] = protocol
        _providerUri = None
        if not json_data['configuration'].get('providerId', None):
            if protocol is not None:
                if protocol == "OIDC10":
                    _providerUri = "/oidc/rp/" + json_data['name']
                    json_data['configuration']['providerId'] = json_data['configuration']['pointOfContactUrl'] + _providerUri
                elif protocol == "SAML2_0":
                    _providerUri = "/" + json_data['name'] + "/saml20"
                    json_data['configuration']['providerId'] = json_data['configuration']['pointOfContactUrl'] + _providerUri
                elif protocol == "WSFED":
                    #  _providerUri = "/wsfed/wsf" #not correct
                    logger.debug("Ignore wsfed for this")

        if protocol == "SAML2_0" and role == 'sp':
            if not json_data['configuration'].get('manageNameIDService', None):
                json_data['configuration']['manageNameIDService'] = []
            if not json_data['configuration'].get('sessionStateHeaders', None):
                json_data['configuration']['sessionStateHeaders'] = []
            if not json_data['configuration'].get('attributeMapping', None):
                json_data['configuration']['attributeMapping'] = {"map": []}
            if not json_data['configuration'].get('wayfCookieLifetime', None):
                json_data['configuration']['wayfCookieLifetime'] = 86400
            if json_data['configuration'].get('authnReqMapping', None):
                # remove javascript
                if json_data['configuration']['authnReqMapping'].get('properties', None):
                    if json_data['configuration']['authnReqMapping']['properties'].get('ruleType', None):
                        del json_data['configuration']['authnReqMapping']['properties']['ruleType']
            else:
                # default
                if tools.version_compare(isamAppliance.facts["version"], "10.0.3") >= 0:
                    json_data['configuration']['authnReqMapping'] = {"activeDelegateId": "skip-authn-request-map"}
            if json_data['configuration'].get('extensionMapping', None):
                # remove javascript
                if json_data['configuration']['extensionMapping'].get('properties', None):
                    if json_data['configuration']['extensionMapping']['properties'].get('ruleType', None):
                        del json_data['configuration']['extensionMapping']['properties']['ruleType']
            else:
                if tools.version_compare(isamAppliance.facts["version"], "10.0.1") >= 0:
                    json_data['configuration']['extensionMapping'] = {"activeDelegateId": "skip-extension-map"}
            # artifactResolutionService bindings url
            if json_data['configuration'].get('artifactResolutionService', None):
                for i in json_data['configuration']['artifactResolutionService']:
                    if not i.get('url', None):
                        # put in a default value : point of contact + federation
                        i['url'] = f"{json_data['configuration']['providerId']}/soap"

            # assertionConsumerService binding url
            if json_data['configuration'].get('assertionConsumerService', None):
                for i in json_data['configuration']['assertionConsumerService']:
                    if not i.get('url', None):
                        # put in a default value : point of contact + federation
                        i['url'] = f"{json_data['configuration']['providerId']}/login"
            # isamAppliance.logger.debug('singlelogoutblabla')
            if json_data['configuration'].get('singleLogoutService', None):
                for i in json_data['configuration']['singleLogoutService']:
                    if not i.get('url', None):
                        # put in a default value : point of contact + federation
                        i['url'] = f"{json_data['configuration']['providerId']}/slo"
            # nameid supported
            if json_data['configuration'].get('nameIDFormat', None):
                if not json_data['configuration']['nameIDFormat'].get('supported', None):
                    if ret_obj['data']['configuration']['nameIDFormat'].get('supported', None):
                        logger.debug("Ignoring supported nameIDFormat for comparison")
                        ret_obj['data']['configuration']['nameIDFormat'].pop('supported')

        sorted_json_data = json.dumps(json_data, skipkeys=True, sort_keys=True)
        # isamAppliance.logger.debug(f"\nSorted Desired:\n\n {sorted_json_data}\n")
        logger.debug(f"\nSorted Desired:\n\n {sorted_json_data}\n")
        sorted_ret_obj = json.dumps(ret_obj['data'], skipkeys=True, sort_keys=True)
        # isamAppliance.logger.debug(f"\nSorted Current:\n\n {sorted_ret_obj}\n")
        logger.debug(f"\nSorted Desired:\n\n {sorted_json_data}\n")
        if sorted_ret_obj != sorted_json_data:
            # parameters that are necessary for compare, but not for update
            json_data.pop('protocol')
            isamAppliance.logger.info("Changes detected, update needed.")
            update_required = True

    return fed_id, update_required, json_data


def export_metadata(isamAppliance, name, filename, check_mode=False, force=False):
    """
    Export a federation
    """
    ret_obj = search(isamAppliance, name)
    fed_id = ret_obj['data']
    if fed_id == {}:
        from ibmsecurity.appliance.ibmappliance import IBMError
        raise IBMError("999", "Cannot export data from unknown federation: {0}".format(name))
    import os.path

    if force is True or (fed_id != {} and os.path.exists(filename) is False):
        if check_mode is False:  # No point downloading a file if in check_mode
            return isamAppliance.invoke_get_file(
                "Export a federation",
                "{0}/{1}/metadata".format(uri, fed_id),
                filename,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def search(isamAppliance, name, force=False, check_mode=False):
    """
    Search federation ID by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['name'] == name:
            logger.info("Found federation {0} id: {1}".format(name, obj['id']))
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Federations between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['id']
    for obj in ret_obj2['data']:
        del obj['id']

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id'])



import logging
from abc import ABCMeta, abstractmethod


class IBMError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class IBMFatal(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class IBMResponse(dict):
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def succeeded_with_data(self):
        """
        Determines whether the execution succeeded with data retrieved.
        :return: True if the execution succeeded and the data is retrieved.
        """
        if self.get('rc', -1) == 0 and self.get("data"):
            return True
        return False

    def succeeded(self):
        """
        Determines whether the execution succeeded.
        :return: True if succeeded.
        """
        if self.get('rc', -1) == 0:
            return True
        return False

    def failed(self):
        """
        Determines whether the execution failed.
        :return: True if the execution failed.
        """
        if self.get('rc', -1) == 0:
            return False
        return True


class IBMAppliance(metaclass=ABCMeta):

    def __init__(self, hostname, user):
        self.logger = logging.getLogger(__name__)
        self.logger.debug('Creating an IBMAppliance')

        self.hostname = hostname
        self.user = user

        self.facts = {}
        self.get_facts()

    @abstractmethod
    def invoke_post_files(self, description, uri, fileinfo, data, ignore_error=False):
        """
        Send multipart/form-data upload file request to the appliance.
        """
        pass

    @abstractmethod
    def invoke_get_file(self, description, uri, filename, ignore_error=False):
        """
        Invoke a GET request and download the response data to a file
        """

    @abstractmethod
    def invoke_put(self, description, uri, data, ignore_error=False):
        """
        Send a PUT request to the LMI.
        """
        pass

    @abstractmethod
    def invoke_post(self, description, uri, data, ignore_error=False):
        """
        Send a POST request to the LMI.
        """
        pass

    @abstractmethod
    def invoke_get(self, description, uri, ignore_error=False):
        """
        Send a GET request to the LMI.
        """
        pass

    @abstractmethod
    def invoke_delete(self, description, uri, ignore_error=False):
        """
        Send a DELETE request to the LMI.
        """
        pass

    @abstractmethod
    def get_facts(self):
        """
        Extracts standard facts from the appliance

        Store it in JSON variable called "facts"
        """
        pass

    def create_return_object(self, rc=0, data={}, warnings=[],
                             changed=False):
        """
        Create a response object with the given properties.
        :param rc: The return code of the call.
        :param data: the data object of the response. Often in Json.
        :param warnings: The warnings of the executed call.
        :param changed: Whether there was any change.
        :return: The IBMResponse object.
        """
        return IBMResponse({'rc': rc,
                            'data': data,
                            'changed': changed,
                            'warnings': warnings,
                            'status_code': 0
                           })



import logging
from ibmsecurity.isam.web.reverse_proxy import junctions

logger = logging.getLogger(__name__)
requires_modules = ["wga"]
requires_version = "9.0.6.0"


def config(
    isamAppliance,
    instance_id,
    hostname="127.0.0.1",
    port=443,
    username=None,
    password=None,
    junction="/mga",
    reuse_certs=False,
    reuse_acls=False,
    check_mode=False,
    force=False,
):
    """
    Authentication and Context based access configuration for a reverse proxy instance

    :param isamAppliance:
    :param instance_id:
    :param junction:
    :param hostname:
    :param port:
    :param username:
    :param password:
    :param reuse_certs:
    :param reuse_acls:
    :param check_mode:
    :param force:
    :return:
    """
    if username is None:
        logger.info("Required parameter username missing. Skipping config.")
        return isamAppliance.create_return_object(
            warning=["Required parameter username missing. Skipping config."]
        )

    if password is None:
        logger.info("Required parameter password missing. Skipping config.")
        return isamAppliance.create_return_object(
            warning=["Required parameter password missing. Skipping config."]
        )
    warnings = [
        "Idempotency logic will check for existence of {} junction. Use force=True to override.".format(
            junction
        )
    ]
    if force is True or _check_config(isamAppliance, instance_id, junction) is False:
        json_data = {
            "junction": junction,
            "hostname": hostname,
            "port": port,
            "username": username,
            "password": password,
            "reuse_certs": reuse_certs,
            "reuse_acls": reuse_acls,
        }
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post(
                " Authentication and Context based access configuration for a reverse proxy instance",
                "/wga/reverseproxy/{}/authsvc_config".format(instance_id),
                json_data,
                warnings=warnings,
                requires_modules=requires_modules,
                requires_version=requires_version,
            )

    return isamAppliance.create_return_object(warnings=warnings)


def _check_config(isamAppliance, instance_id, junction):
    """
    Check if the junction for aac already created

    :param isamAppliance:
    :param instance_id:
    :param junction:
    :return:
    """
    ret_obj = junctions.get_all(isamAppliance, instance_id)

    for j in ret_obj["data"]:
        if j["id"] == junction:
            logger.info(
                "Junction {} was found - hence aac config must have already executed.".format(
                    junction
                )
            )
            return True

    return False



import logging

import json
from ibmsecurity.utilities.tools import version_compare
from ibmsecurity.utilities.tools import json_compare

logger = logging.getLogger(__name__)

# URI for this module
uri = "/mga/server_connections/redis"
requires_modules = ["mga", "federation"]
requires_version = "10.0.3.0"  # Will change if introduced in an earlier version.

def get_all(isamAppliance, check_mode=False, force=False):
   """
   Retrieving a list of all Redis connections
   """
   return isamAppliance.invoke_get("Retrieving a list of all Redis connections",
                                   "{0}/v1".format(uri), requires_modules=requires_modules,
                                   requires_version=requires_version)


def get(isamAppliance, name=None, check_mode=False, force=False):
   """
   Retrieving a Redis connection
   """
   ret_obj = search(isamAppliance, name=name, force=force)
   id = ret_obj["data"]

   if id == {}:
       logger.info("Redis Service connection {0} had no match, skipping retrieval.".format(name))
       return isamAppliance.create_return_object()
   else:
       return isamAppliance.invoke_get("Retrieving a Redis Service connection",
                                       "{0}/{1}/v1".format(uri, id),
                                       requires_modules=requires_modules,
                                       requires_version=requires_version)


def search(isamAppliance, name, check_mode=False, force=False):
   """
   Search UUID for named Redis Service connection
   """
   ret_obj = get_all(isamAppliance)
   return_obj = isamAppliance.create_return_object()
   return_obj["warnings"] = ret_obj["warnings"]

   for obj in ret_obj['data']:
       if obj['name'] == name:
           logger.info("Found Redis Service connection {0} id: {1}".format(name, obj['uuid']))
           return_obj['data'] = obj['uuid']
           return_obj['rc'] = 0

   return return_obj


def add(isamAppliance, name, connection, description='', connectionManager=None, servers=None, locked=False, check_mode=False, force=False):
   """
   Creating a Redis Service connection
   """
   if (search(isamAppliance, name=name))['data'] == {}:
       if check_mode is True:
           return isamAppliance.create_return_object(changed=True)
       else:
           return isamAppliance.invoke_post(
               "Creating a Redis Service connection",
               "{0}/v1".format(uri),
               _create_json(name=name,
                            description=description,
                            locked=locked,
                            connection=connection,
                            connectionManager=connectionManager,
                            servers=servers),
               requires_modules=requires_modules, requires_version=requires_version)

   return isamAppliance.create_return_object()


def delete(isamAppliance, name, check_mode=False, force=False):
   """
   Deleting a Redis connection
   """
   ret_obj = search(isamAppliance, name, check_mode=check_mode, force=force)
   id = ret_obj["data"]

   if force is True or _check_exists(isamAppliance, name=name) is True:
       if check_mode is True:
           return isamAppliance.create_return_object(changed=True)
       else:
           return isamAppliance.invoke_delete(
               "Deleting a Redis connection",
               "{0}/{1}/v1".format(uri, id), requires_modules=requires_modules,
               requires_version=requires_version)

   return isamAppliance.create_return_object()


def update(isamAppliance, name, connection, description='', locked=False, connectionManager=None, servers=None, new_name=None, ignore_password_for_idempotency=False,
          check_mode=False, force=False):
   """
   Modifying a Redis connection

   Use new_name to rename the connection
   """
   ret_obj = get(isamAppliance, name)
   warnings = ret_obj["warnings"]

   if ret_obj["data"] == {}:
       warnings.append("Redis connection {0} not found, skipping update.".format(name))
       return isamAppliance.create_return_object(warnings=warnings)
   else:
       id = ret_obj["data"]["uuid"]

   needs_update = False

   json_data = _create_json(name=name,
                            description=description,
                            locked=locked,
                            connection=connection,
                            connectionManager=connectionManager,
                            servers=servers)
   if new_name is not None:  # Rename condition
       json_data['name'] = new_name

   if force is not True:
       if 'uuid' in ret_obj['data']:
           del ret_obj['data']['uuid']
       if ignore_password_for_idempotency:
           if 'password' in connection:
               warnings.append("Request made to ignore password for idempotency check.")
               connection.pop('password', None)

       #remove uuid from servers
       if 'servers' in ret_obj['data']:
         for x in ret_obj['data']['servers']:
             if 'uuid' in x:
                 logger.debug("Deleting uuid from returned object:\n{0}".format(x['uuid']))
                 del x['uuid']

       sorted_ret_obj = json.dumps(ret_obj['data'], skipkeys=True, sort_keys=True)
       sorted_json_data = json.dumps(json_data, skipkeys=True, sort_keys=True)

       logger.debug("Sorted Existing Data:\n{0}".format(sorted_ret_obj))
       logger.debug("Sorted Desired  Data:\n{0}".format(sorted_json_data))

       if sorted_ret_obj != sorted_json_data:
           needs_update = True

       if 'password' in connection:
           warnings.append("Since existing password cannot be read - this call will not be idempotent.")
           needs_update = True

   if force is True or needs_update:
       if check_mode:
           return isamAppliance.create_return_object(changed=True, warnings=warnings)
       else:
           return isamAppliance.invoke_put(
               "Modifying a Redis connection",
               "{0}/{1}/v1".format(uri, id), json_data, requires_modules=requires_modules,
               requires_version=requires_version, warnings=warnings)

   return isamAppliance.create_return_object(warnings=warnings)

def set(isamAppliance, name, connection, description='', locked=False, new_name=None, connectionManager=None, servers=None, ignore_password_for_idempotency=False, check_mode=False, force=False):
   """
   Creating or Modifying a Redis connection
   """
   if (search(isamAppliance, name=name))['data'] == {}:
       # Force the add - we already know connection does not exist
       return add(isamAppliance=isamAppliance, name=name, connection=connection, description=description,
                  locked=locked, connectionManager=connectionManager, servers=servers, check_mode=check_mode, force=True)
   else:
       # Update request
       return update(isamAppliance=isamAppliance, name=name, connection=connection, description=description,
                     locked=locked, connectionManager=connectionManager, servers=servers,
                     new_name=new_name, ignore_password_for_idempotency=ignore_password_for_idempotency,
                     check_mode=check_mode, force=force)

def _create_json(name, description, locked, connection, connectionManager=None, servers=None):
   """
   Create a JSON to be used for the REST API call
   """
   _json = {
       "connection": connection,
       "type": "redis",
       "name": name,
       "description": description,
       "locked": locked
   }
   if connectionManager is not None:
       _json['connectionManager'] = connectionManager
   if servers is not None:
       _json['servers'] = servers
   return _json


def compare(isamAppliance1, isamAppliance2):
   """
   Compare Redis connections between two appliances
   """
   ret_obj1 = get_all(isamAppliance1)
   ret_obj2 = get_all(isamAppliance2)

   for obj in ret_obj1['data']:
       del obj['uuid']
   for obj in ret_obj2['data']:
       del obj['uuid']

   return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['uuid'])


def _check_exists(isamAppliance, name=None, id=None):
   """
   Check if Redis Connection already exists
   """
   ret_obj = get_all(isamAppliance)

   for obj in ret_obj['data']:
       if (name is not None and obj['name'] == name) or (id is not None and obj['uuid'] == id):
           return True

   return False

#test connection for redis is different from the other ones (those are in connection.py)
def test(isamAppliance, name, check_mode=False, force=False):
   """
   Test Redis Connection
   Note that to be able to test the connection, the LMI needs to be restarted!
   Otherwise, you'll run into : FBTSPS137E The Redis configuration is invalid and cannot be processed
   param: name:    the name of the server_connection
   """
   if check_mode is True:
       return isamAppliance.create_return_object()

   return isamAppliance.invoke_get("Test Redis Connection", f"/iam/access/v8/testconnection/redis/{name}")


import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get(isvgAppliance, check_mode=False, force=False):
    """
    Get management authentication
    """
    return isvgAppliance.invoke_get("Get management authentication",
                                    "/lmi_auth/")


def set(isvgAppliance,
        ldap_host=None,
        ldap_port=None,
        base_dn=None,
        admin_group_dn=None,
        enable_ssl=False,
        user_attribute='uid',
        group_member_attribute='member',
        bind_dn=None,
        bind_password=None,
        check_mode=False,
        force=False):
    """
    Set management authentication to remote
    """
    warnings = []
    update_required = False
    json_data = {
    }
    if ldap_host is not None:
        json_data['ldap_host'] = ldap_host
    if ldap_port is not None:
        json_data['ldap_port'] = ldap_port
    if enable_ssl is not None:
        json_data['enable_ssl'] = enable_ssl
    if user_attribute is not None:
        json_data['user_attribute'] = user_attribute
    if group_member_attribute is not None:
        json_data['group_member_attribute'] = group_member_attribute
    if base_dn is not None:
        json_data['base_dn'] = base_dn
    if admin_group_dn is not None:
        json_data['admin_group_dn'] = admin_group_dn
    if bind_dn is not None:
        json_data['bind_dn'] = bind_dn
    if bind_password is not None:
        json_data['bind_password'] = bind_password
    if force is False:
        if bind_password is not None:
            warnings.append("Unable to read existing bind password to check idempotency.")
            update_required = True
        else:
            ret_obj = get(isvgAppliance)
            if 'bind_dn' in ret_obj['data']:
                if ret_obj['data']['bind_dn'] is None:
                    del ret_obj['data']['bind_dn']
            sorted_json_data = ibmsecurity.utilities.tools.json_sort(json_data)
            logger.debug("Sorted input: {0}".format(sorted_json_data))
            sorted_ret_obj = ibmsecurity.utilities.tools.json_sort(ret_obj['data'])
            logger.debug("Sorted existing data: {0}".format(sorted_ret_obj))
            if sorted_ret_obj != sorted_json_data:
                logger.info("Changes detected, update needed.")
                update_required = True

    if force is True or update_required is True:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isvgAppliance.invoke_put(
                "Set management authentication to remote",
                "/lmi_auth/",
                json_data, warnings=warnings)

    return isvgAppliance.create_return_object()


def disable(isvgAppliance, check_mode=False, force=False):
    """
    Disable remote management authentication
    """
    return set(isvgAppliance=isvgAppliance, ldap_host=None, ldap_port=None, base_dn=None, admin_group_dn=None,
               enable_ssl=None, user_attribute=None,
               group_member_attribute=None, bind_dn=None, bind_password=None, check_mode=check_mode,
               force=force)


def test(isvgAppliance, userid, password, check_mode=False, force=False):
    """
    Testing the management authentication
    """
    ret_obj = isvgAppliance.invoke_post("Testing the management authentication",
                                        "/lmi_auth/",
                                           {
                                            'user': userid,
                                            'password': password
                                           }
                                       )
    ret_obj['changed'] = False

    return ret_obj


def compare(isvgAppliance1, isvgAppliance2):
    """
    Compare management authentication settings
    """
    ret_obj1 = get(isvgAppliance1)
    ret_obj2 = get(isvgAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])